import numpy as np
from xbbo.core.trials import Trial, Trials
from xbbo.initial_design import ALL_avaliable_design
from xbbo.search_algorithm.base import AbstractOptimizer
from . import alg_register


@alg_register.register('rs')
class RandomOptimizer(AbstractOptimizer):
    def __init__(
            self,
            space,
            seed: int = 42,
            initial_design: str = 'random',
            #  min_sample=1,
            suggest_limit: int = np.inf,
            **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='round',
                                   encoding_ord='round',
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)
        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget
        self.initial_design_configs = self.initial_design.select_configurations(
        )
        self.dimension = self.space.get_dimensions()
        self.trials = Trials(dim=self.dimension)

    def _suggest(self, n_suggestions=1):
        trial_list = []
        for n in range(n_suggestions):
            if self.initial_design_configs:
                config = self.initial_design_configs.pop(0)
                trial_list.append(
                    Trial(
                        configuration=config,
                        config_dict=config.get_dictionary(),
                        #   array=config.get_array(sparse=False))
                    ))
                continue
            iter_ = 0
            while iter_ < 1000:  # remove history suggest
                config = self.space.sample_configuration()[0]
                if not self.trials.is_contain(config):
                    trial_list.append(
                        Trial(configuration=config,
                              config_dict=config.get_dictionary(),
                              array=config.get_array(sparse=False)))

                    break
                iter_ += 1
            else:
                assert False, "no more configs can be suggest"

        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)


opt_class = RandomOptimizer


from abc import ABC, abstractmethod
import numpy as np

class AbstractBaseModel(ABC):

    def __init__(self, **params):
        self.params = params
        pass

    @abstractmethod
    def fit(self, dataset):
        pass

    @abstractmethod
    def get_score(self, test_data, metric):
        '''
        score 指标 越大越好
        '''
        pass



class TestFunction(ABC):
    """Abstract base class for test functions in the benchmark. These do not need to be ML hyper-parameter tuning.
    """

    def __init__(self, seed=42):
        """Setup general test function for benchmark. We assume the test function knows the meta-data about the search
        configspace, but is also stateless to fit modeling assumptions. To keep stateless, it does not do things like count
        the number of function evaluations.
        """
        # This will need to be set before using other routines
        self.api_config = None
        self.rng = np.random.RandomState(seed)

    @abstractmethod
    def evaluate(self, params):
        """Abstract method to evaluate the function at a parameter setting.
        """

    def get_api_config(self):
        """Get the API config for this test problem.

        Returns
        -------
        api_config : dict(str, dict(str, object))
            The API config for the used model. See README for API description.
        """
        assert self.api_config is not None, "API config is not set."
        assert isinstance(self.api_config, dict), "API config is not a dict type"
        return self.api_config

    def _load_data(self, dataset, data_root=None):
        pass

    def _load_api_config(self):
        pass

    def _load_base_model(self):
        '''
        return a class
        '''
        pass


import math
from typing import Optional, List, Tuple, cast

import numpy as np

from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trial, Trials
from . import alg_register

@alg_register.register('rea')
class RegularizedEA(AbstractOptimizer):
    '''
    Regularized Evolution for Image Classifier Architecture Search
    '''

    def __init__(self,
                 space: DenseConfigurationSpace,
                 seed: int = 42,
                 llambda=None,
                 sample_size=None,
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)
        # FeatureSpace_gaussian.__init__(self, self.space.dtypes_idx_map)
        # Uniform2Gaussian.__init__(self)
        # configs = self.space.get_hyperparameters()
        self.dimension = self.space.get_dimensions()
        self.bounds = self.space.get_bounds()
        # self.num_of_tour_particips = kwargs.get('n_part',2)
        self.llambda = llambda if llambda else 4 + math.floor(3 * math.log(self.dimension)) # (eq. 48)
        self.tournament_sample_size = self.llambda//2 if sample_size is None else min(max(sample_size, 1), self.llambda)
        self.population = self.create_initial_population()
        self.population_y = None

        # self.mu = kwargs.get('mu',20) # 交叉和变异算法的分布指数
        self.mum = kwargs.get('mum',20)
        # self.crossrate = kwargs.get('crossrate',0.9)
        # ---



        self.trials = Trials(dim=self.dimension)
        self.cur = 0
        self.gen = 0
        self.listy = []

    def _suggest(self, n_suggestions=1):
        assert self.llambda % n_suggestions == 0
        trial_list = []
        for n in range(n_suggestions):
            new_individual = self.population[self.cur]
            new_individual = np.clip(new_individual, self.bounds.lb, self.bounds.ub)
            # array = self.feature_to_array(new_individual)
            config = DenseConfiguration.from_array(self.space,new_individual)
            self.cur += 1
            trial_list.append(
            Trial(config,
                    config_dict=config.get_dictionary(),
                    array=new_individual,
                    origin='REA', loc=self.cur))
        

        # self._num_suggestions += n_suggestions
        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)
            self.listy.append(trial.observe_value)
        if self.cur == len(self.population):
            if self.population_y is None:
                self.population_y = np.asarray(self.listy)
            else:
                self.population_y = np.concatenate([self.population_y, self.listy], axis=0)

            # remove dead
            self.population_y = np.asarray(self.population_y[-self.llambda:])
            self.population = np.asarray(self.population[-self.llambda:])

            s_id = np.argsort(self.population_y)
            rank = np.argsort(s_id)
            parents_id = self.rng.choice(len(rank),replace=True, size=self.tournament_sample_size)
            parent_id = s_id[rank[parents_id].min()]

            children = self.__mutate2(self.population[parent_id])

            # ---
            self.population = np.append(self.population, [children], axis=0)
            self.cur = len(self.population_y)
            assert self.cur < len(self.population)

            self.listy = []
            self.gen += 1


    def calculate_crowding_distance(self, points, m, M):
        cd = np.zeros(len(points))
        if len(cd) <= points.shape[-1]:
            return cd
        for n in range(points.shape[-1]):
            s_id = np.argsort(points[:, n])
            cd[s_id[[0, -1]]] = [np.inf, np.inf]
            for idx in range(len(s_id[1:-1])):
                cd[s_id[idx]] += (points[s_id[idx+1], n] - points[s_id[idx-1], n]) / (M[n] - m[n])
        return cd



    def create_initial_population(self):
        # return self.rng.normal(0, 1, size=(self.llambda, self.dimension))
        return self.rng.uniform(0, 1, size=(self.llambda, self.dimension))


    def __mutate2(self, parent):
        child = parent.copy()
        r = self.rng.rand(len(child))
        delta = np.where(r<0.5, (2 * r) ** (1 / (self.mum + 1)) - 1, 1 - (2*(1 - r))**(1/(self.mum+1)))
        return np.clip(child + delta, self.bounds.lb, self.bounds.ub)
    
    def __mutate_naive(self, parent):
        child = parent.copy()
        child[self.rng.randint(self.dimension)] = self.rng.uniform(0, 1)
        return child

    def __mutate(self, parent, mu0=0.7):
        if self.rng.rand() <= mu0:
            return self.__mutate_naive(parent)
        else:
            return self.rng.uniform(0, 1, size=self.dimension)


opt_class = RegularizedEA



from collections import OrderedDict
import numpy as np
import abc
from typing import Iterable, List, Tuple, Union
# import logging

from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace, convert_denseConfigurations_to_array
from xbbo.core.trials import Trials
from xbbo.surrogate.base import SurrogateModel


class AbstractAcquisitionFunction(object, metaclass=abc.ABCMeta):
    """Abstract base class for acquisition function

    Attributes
    ----------
    model
    logger
    """

    # def __str__(self):
    #     return type(self).__name__ + " (" + self.long_name + ")"

    def __init__(self, surrogate_model: Union[SurrogateModel,
                                              List[SurrogateModel]], **kwargs):
        """Constructor

        Parameters
        ----------
        model : AbstractEPM
            Models the objective function.
        """
        self.surrogate_model = surrogate_model
        # self.logger = logging.getLogger(
        #     self.__module__ + "." + self.__class__.__name__)

    def update(self, **kwargs):
        """Update the acquisition functions values.

        This method will be called if the surrogate is updated. E.g.
        entropy search uses it to update its approximation of P(x=x_min),
        EI uses it to update the current optimizer.

        The default implementation takes all keyword arguments and sets the
        respective attributes for the acquisition function object.

        Parameters
        ----------
        kwargs
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __call__(self,
                 configurations: Union[List[DenseConfiguration], np.ndarray],
                 convert=True,
                 **kwargs):
        """Computes the acquisition value for a given X

        Parameters
        ----------
        configurations : list
            The configurations where the acquisition function
            should be evaluated.
        convert : bool

        Returns
        -------
        np.ndarray(N, 1)
            acquisition values for X
        """
        if convert:
            X = convert_denseConfigurations_to_array(configurations)
        else:
            X = configurations  # to be compatible with multi-objective acq to call single acq
        if len(X.shape) == 1:
            X = X[np.newaxis, :]

        acq = self._compute(X, **kwargs)
        if np.any(np.isnan(acq)):
            idx = np.where(np.isnan(acq))[0]
            acq[idx, :] = -np.finfo(np.float).max
        return acq

    @abc.abstractmethod
    def _compute(self, X: np.ndarray, **kwargs):
        """Computes the acquisition value for a given point X. This function has
        to be overwritten in a derived class.

        Parameters
        ----------
        X : np.ndarray
            The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        Returns
        -------
        np.ndarray(N,1)
            Acquisition function values wrt X
        """
        raise NotImplementedError()
    
    def update_weight(self, w, rho=None):
        pass


class AcquisitionFunctionMaximizer(object, metaclass=abc.ABCMeta):
    """Abstract class for acquisition maximization.

    In order to use this class it has to be subclassed and the method
    ``_maximize`` must be implemented.

    Parameters
    ----------
    acquisition_function : ~xbbo.acquisition_function.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.config_space.ConfigurationSpace

    rng : np.random.RandomState or int, optional
    """
    def __init__(self,
                 acquisition_function: AbstractAcquisitionFunction,
                 config_space: DenseConfigurationSpace,
                 rng: np.random.RandomState = np.random.RandomState(42)):
        self.acquisition_function = acquisition_function
        self.config_space = config_space
        # self.logger = logging.getLogger(self.__module__ + "." +
                                        # self.__class__.__name__)
        self.rng = rng

    def maximize(self,
                 trials: Trials,
                 num_points: int,
                 drop_self_duplicate: bool = False,
                 **kwargs) -> Iterable[DenseConfiguration]:
        """Maximize acquisition function using ``_maximize``.

        Parameters
        ----------
        trials: ~xbbo.utils.history_container.HistoryContainer
            trials object
        stats: ~xbbo.stats.stats.Stats
            current stats object
        num_points: int
            number of points to be sampled
        **kwargs

        Returns
        -------
        iterable
            An iterable consisting of :class:`xbbo.config_space.DenseConfiguration`.
        """
        configs = [t[1] for t in self._maximize(trials, num_points, **kwargs)]
        return self.unique(configs=configs) if drop_self_duplicate else configs

    @staticmethod
    def unique(configs: Iterable[DenseConfiguration]):
        return list(OrderedDict.fromkeys(configs))

    @abc.abstractmethod
    def _maximize(self, trials: Trials, num_points: int,
                  **kwargs) -> Iterable[Tuple[float, DenseConfiguration]]:
        """Implements acquisition function maximization.

        In contrast to ``maximize``, this method returns an iterable of tuples,
        consisting of the acquisition function value and the DenseConfiguration. This
        allows to plug together different acquisition function maximizers.

        Parameters
        ----------
        trials: ~xbbo.utils.history_container.HistoryContainer
            trials object
        stats: ~xbbo.stats.stats.Stats
            current stats object
        num_points: int
            number of points to be sampled
        **kwargs

        Returns
        -------
        iterable
            An iterable consistng of
            tuple(acqusition_value, :class:`xbbo.config_space.DenseConfiguration`).
        """
        raise NotImplementedError()

    def _sort_configs_by_acq_value(
        self, configs: List[DenseConfiguration]
    ) -> List[Tuple[float, DenseConfiguration]]:
        """Sort the given configurations by acquisition value

        Parameters
        ----------
        configs : list(DenseConfiguration)

        Returns
        -------
        list: (acquisition value, Candidate solutions),
                ordered by their acquisition function value
        """

        acq_values = self.acquisition_function(configs)

        # From here
        # http://stackoverflow.com/questions/20197990/how-to-make-argsort-result-to-be-random-between-equal-values
        random = self.rng.rand(len(acq_values))
        # Last column is primary sort key!
        indices = np.lexsort((random.flatten(), acq_values.flatten()))

        # Cannot use zip here because the indices array cannot index the
        # rand_configs list, because the second is a pure python list
        # https://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-whilst-preserving-order
        # seen = set()
        # seen_add = seen.add

        return [(acq_values[ind][0], configs[ind]) for ind in indices[::-1]]



import logging, typing
import numpy as np
from xbbo.acquisition_function.transfer.mogp import MoGP_AcqFunc
from xbbo.acquisition_function.transfer.taf import TAF_AcqFunc
from xbbo.acquisition_function.acq_func import EI_AcqFunc
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.surrogate.gaussian_process import GPR_sklearn
from . import alg_register
from xbbo.core.trials import Trial, Trials
from xbbo.initial_design import ALL_avaliable_design
from xbbo.acquisition_function.acq_optimizer import InterleavedLocalAndRandomSearch, LocalSearch, RandomScipyOptimizer, RandomSearch, ScipyGlobalOptimizer, ScipyOptimizer
from xbbo.surrogate.transfer.weight_stategy import KernelRegress, RankingWeight, ZeroWeight
from xbbo.surrogate.transfer.tst import BaseModel, TST_surrogate

logger = logging.getLogger(__name__)


@alg_register.register('bo-transfer')
class SMBO(AbstractOptimizer):
    def __init__(self,
                 space: DenseConfigurationSpace,
                 seed: int = 42,
                 initial_design: str = 'sobol',
                 suggest_limit: int = np.inf,
                 surrogate: str = 'gp',
                 acq_func: str = 'ei',
                 acq_opt: str = 'rs_ls',
                 predict_x_best: bool = False,
                 weight_srategy: str = 'kernel',
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)
        self.predict_x_best = predict_x_best
        self.dimension = self.space.get_dimensions()

        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget
        self.hp_num = len(self.space)
        self.initial_design_configs = self.initial_design.select_configurations(
        )
        self.trials = Trials(dim=self.dimension)

        # self.rho = kwargs.get("rho", 1)
        self.bandwidth = kwargs.get("bandwdth", 0.1)
        self.base_models = kwargs.get("base_models")
        if self.base_models:
            assert isinstance(self.base_models[0], BaseModel)
            if surrogate == 'gp':
                self.surrogate_model = GPR_sklearn(self.space, rng=self.rng)
            elif surrogate == 'tst':
                self.surrogate_model = TST_surrogate(self.space,
                                                     self.base_models,
                                                     rng=self.rng)
        else:
            raise NotImplementedError()
        if weight_srategy == 'kernel':
            self.weight_sratety = KernelRegress(self.space, self.base_models,
                                                self.surrogate_model, self.rng)
        elif weight_srategy == 'rw':
            self.weight_sratety = RankingWeight(self.space,
                                                self.base_models,
                                                self.surrogate_model,
                                                self.rng,
                                                budget=suggest_limit,
                                                is_purn=True)
        elif weight_srategy == 'zero':
            self.weight_sratety = ZeroWeight(self.space, self.base_models,
                                             self.surrogate_model, self.rng)
        else:
            raise NotImplementedError()

        if acq_func == 'mogp':
            self.acquisition_func = MoGP_AcqFunc(self.surrogate_model,
                                                 self.base_models, self.rng)
        elif acq_func == 'taf':
            self.acquisition_func = TAF_AcqFunc(self.surrogate_model,
                                                self.base_models, self.rng)
        elif acq_func == 'ei':
            self.acquisition_func = EI_AcqFunc(self.surrogate_model, self.rng)
        else:
            raise NotImplementedError()

        if acq_opt == 'ls':
            self.acq_maximizer = LocalSearch(self.acquisition_func, self.space,
                                             self.rng)
        elif acq_opt == 'rs':
            self.acq_maximizer = RandomSearch(self.acquisition_func,
                                              self.space, self.rng)
        elif acq_opt == 'rs_ls':
            self.acq_maximizer = InterleavedLocalAndRandomSearch(
                self.acquisition_func, self.space, self.rng)
        elif acq_opt == 'scipy':
            self.acq_maximizer = ScipyOptimizer(self.acquisition_func,
                                                self.space, self.rng)
        elif acq_opt == 'scipy_global':
            self.acq_maximizer = ScipyGlobalOptimizer(self.acquisition_func,
                                                      self.space, self.rng)
        elif acq_opt == 'r_scipy':
            self.acq_maximizer = RandomScipyOptimizer(self.acquisition_func,
                                                      self.space, self.rng)
        else:
            raise ValueError('acq_opt {} not in {}'.format(
                acq_opt,
                ['ls', 'rs', 'rs_ls', 'scipy', 'scipy_global', 'r_scipy']))
        logger.info(
            "Execute Bayesian optimization...\n [Using ({})surrogate, ({})acquisition function, ({})acquisition optmizer]"
            .format(surrogate, acq_func, acq_opt))

    def _suggest(self, n_suggestions=1):
        trial_list = []
        # currently only suggest one
        if (self.trials.trials_num) < self.init_budget:
            assert self.trials.trials_num % n_suggestions == 0
            configs = self.initial_design_configs[
                int(n_suggestions *
                    self.trials.trials_num):int(n_suggestions *
                                                (self.trials.trials_num + 1))]
            for config in configs:
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=config.get_array(sparse=False)))
        else:
            # update target surrogate model
            self.surrogate_model.train(
                np.asarray(self.trials.get_array()),
                np.asarray(self.trials.get_history()[0]))
            # calculate base incuments (only use for acq base EI)
            observed_X = self.trials.get_array()
            base_incuments = []
            for model in self.base_models:  # TODO make sure untransform ?
                base_incuments.append(model.predict(observed_X, None)[0].min())
            _, best_val = self._get_x_best(self.predict_x_best)
            self.acquisition_func.update(surrogate_model=self.surrogate_model,
                                         y_best=best_val,
                                         _base_incuments=base_incuments)
            # caculate weight for base+target model
            weight = self.weight_sratety.get_weight(self.trials)
            self.surrogate_model.update_weight(weight)
            self.acquisition_func.update_weight(weight)
            # acq maximize
            configs = []
            configs = self.acq_maximizer.maximize(self.trials,
                                                  1000,
                                                  drop_self_duplicate=True,
                                                  _sorted=True)
            _idx = 0
            for n in range(n_suggestions):
                while _idx < len(configs):  # remove history suggest
                    if not self.trials.is_contain(configs[_idx]):
                        config = configs[_idx]
                        configs.append(config)
                        trial_list.append(
                            Trial(configuration=config,
                                  config_dict=config.get_dictionary(),
                                  array=config.get_array(sparse=False)))
                        _idx += 1

                        break
                    _idx += 1
                else:
                    assert False, "no more configs can be suggest"
                # surrogate = TST_surrogate(self.gps, self.target_model,
                #   self.similarity, self.rho)

        return trial_list

    def _observe(self, trial_list):
        # print(y)
        for trial in trial_list:
            self.trials.add_a_trial(trial)

    def _get_x_best(self, predict: bool) -> typing.Tuple[float, np.ndarray]:
        """Get value, configuration, and array representation of the "best" configuration.

        The definition of best varies depending on the argument ``predict``. If set to ``True``,
        this function will return the stats of the best configuration as predicted by the model,
        otherwise it will return the stats for the best observed configuration.

        Parameters
        ----------
        predict : bool
            Whether to use the predicted or observed best.

        Returns
        -------
        float
        np.ndarry
        Configuration
        """
        if predict:
            X = self.trials.get_array()
            costs = list(
                map(
                    lambda x: (
                        self.surrogate_model.predict(x.reshape((1, -1)))[0][0],
                        x,
                    ),
                    X,
                ))
            costs = sorted(costs, key=lambda t: t[0])
            x_best_array = costs[0][1]
            best_observation = costs[0][0]
            # won't need log(y) if EPM was already trained on log(y)
        else:
            best_idx = self.trials.best_id
            x_best_array = self.trials.get_array()[best_idx]
            best_observation = self.trials.best_observe_value

        return x_best_array, best_observation


opt_class = SMBO



import numpy as np

# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trials, Trial
from . import alg_register


@alg_register.register('de')
class DE(AbstractOptimizer):
    def __init__(self,
                 space: DenseConfigurationSpace,
                 seed: int = 42,
                 llambda=10,
                 boundary_fix_type='random',
                 mutation_factor=0.5,
                 crossover_prob=0.5,
                 strategy='rand1_bin',
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)

        # Uniform2Gaussian.__init__(self,)
        self.dimension = self.space.get_dimensions()
        self.bounds = self.space.get_bounds()
        self.mutation_factor = mutation_factor
        self.crossover_prob = crossover_prob
        self.fix_type = boundary_fix_type
        # self.dimension = len(configs)
        self.strategy = strategy        
        if self.strategy is not None:
            self.mutation_strategy = self.strategy.split('_')[0]
            self.crossover_strategy = self.strategy.split('_')[1]
        else:
            self.mutation_strategy = self.crossover_strategy = None
        self.llambda = max(llambda, self._set_min_pop_size())
        self.population = [None] * self.llambda
        self.population_fitness = [np.inf] * self.llambda
        # self.candidates = [None] * self.llambda
        self.trials = Trials(dim=self.dimension)
        self.current_best = None
        self.current_best_fitness = np.inf
        self._num_suggestions = 0
        # self.F1 = kwargs.get('F1', 0.8)
        # self.F2 = kwargs.get('F2', 0.8)
        # self.CR = kwargs.get('CR', 0.5)

    def _set_min_pop_size(self):
        if self.mutation_strategy in ['rand1', 'rand2dir', 'randtobest1']:
            self._min_pop_size = 3
        elif self.mutation_strategy in ['currenttobest1', 'best1']:
            self._min_pop_size = 2
        elif self.mutation_strategy in ['best2']:
            self._min_pop_size = 4
        elif self.mutation_strategy in ['rand2']:
            self._min_pop_size = 5
        else:
            self._min_pop_size = 1

        return self._min_pop_size

    def fix_boundary(self, individual):
        if self.fix_type == 'random':
            return np.where(
                (individual > self.bounds.lb) & (individual < self.bounds.ub),
                individual,
                self.rng.uniform(self.bounds.lb, self.bounds.ub,
                                 self.dimension))  # FIXME
        elif self.fix_type == 'clip':
            return np.clip(individual, self.bounds.lb, self.bounds.ub)

    def _suggest(self, n_suggestions=1):
        trial_list = []
        for n in range(n_suggestions):
            # suggest_array = []
            idx = self._num_suggestions % self.llambda
            target = self.population[idx]
            if target is not None:
                donor = self.mutation(current=target, best=None)
                if donor is None:
                    candidate = self.rng.uniform(self.bounds.lb,
                                                    self.bounds.ub,
                                                    self.dimension)
                else:
                    candidate = self.crossover(target, donor)
            else:
                candidate = self.rng.uniform(self.bounds.lb,
                                                    self.bounds.ub,
                                                    self.dimension)


            self._num_suggestions += 1
            candidate = self.fix_boundary(candidate)
            # array = self.feature_to_array(new_individual)
            config = DenseConfiguration.from_array(self.space, candidate)
            trial_list.append(
                Trial(config,
                      config_dict=config.get_dictionary(),
                      array=candidate,
                      origin='DE',
                      loc=idx))

        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)
            # idx = self.candidates.index(tuple(trial.array))
            idx = trial.loc
            # selection
            if trial.observe_value <= self.population_fitness[idx]:
                self.population[idx] = trial.array
                self.population_fitness[idx] = trial.observe_value
            # self._num_suggestions += 1
            # self.candidates[idx] = None
                if trial.observe_value < self.current_best_fitness:
                    self.current_best = self.population[idx]
                    self.current_best_fitness = trial.observe_value
                
    def _mutation_rand1(self, r1, r2, r3):
        '''Performs the 'rand1' type of DE mutation
        '''
        
        diff = r2 - r3
        mutant = r1 + self.mutation_factor * diff
        return mutant

    def _mutation_rand2(self, r1, r2, r3, r4, r5):
        '''Performs the 'rand2' type of DE mutation
        '''
        diff1 = r2 - r3
        diff2 = r4 - r5
        mutant = r1 + self.mutation_factor * diff1 + self.mutation_factor * diff2
        return mutant

    def _mutation_currenttobest1(self, current, best, r1, r2):
        diff1 = best - current
        diff2 = r1 - r2
        mutant = current + self.mutation_factor * diff1 + self.mutation_factor * diff2
        return mutant

    def _mutation_rand2dir(self, r1, r2, r3):
        diff = r1 - r2 - r3
        mutant = r1 + self.mutation_factor * diff / 2
        return mutant

    def mutation(self, current=None, best=None, alt_pop=None):
        '''Performs DE mutation
        '''
        if self.mutation_strategy == 'rand1':
            r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop)

            mutant = self._mutation_rand1(r1, r2, r3)

        elif self.mutation_strategy == 'rand2':
            r1, r2, r3, r4, r5 = self._sample_population(size=5, alt_pop=alt_pop)
            mutant = self._mutation_rand2(r1, r2, r3, r4, r5)

        elif self.mutation_strategy == 'rand2dir':
            r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop)

            mutant = self._mutation_rand2dir(r1, r2, r3)

        elif self.mutation_strategy == 'best1':
            r1, r2 = self._sample_population(size=2, alt_pop=alt_pop)

            if best is None:
                best = self.population[np.argmin(self.population_fitness)]
            mutant = self._mutation_rand1(best, r1, r2)

        elif self.mutation_strategy == 'best2':
            r1, r2, r3, r4 = self._sample_population(size=4, alt_pop=alt_pop)
            if best is None:
                best = self.population[np.argmin(self.population_fitness)]
            mutant = self._mutation_rand2(best, r1, r2, r3, r4)

        elif self.mutation_strategy == 'currenttobest1':
            r1, r2 = self._sample_population(size=2, alt_pop=alt_pop)
            if best is None:
                best = self.population[np.argmin(self.population_fitness)]
            mutant = self._mutation_currenttobest1(current, best, r1, r2)

        elif self.mutation_strategy == 'randtobest1':
            r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop)
            if best is None:
                best = self.population[np.argmin(self.population_fitness)]
            mutant = self._mutation_currenttobest1(r1, best, r2, r3)

        return mutant

    def _sample_population(self, size: int = 3, alt_pop= None):
        '''Samples 'size' individuals

        If alt_pop is None or a list/array of None, sample from own population
        Else sample from the specified alternate population (alt_pop)
        '''
        if isinstance(alt_pop, list) or isinstance(alt_pop, np.ndarray):
            idx = [indv is None for indv in alt_pop]
            if any(idx):
                selection = self.rng.choice(np.arange(len(self.population)), size, replace=False)
                return np.array(self.population)[selection]
            else:
                if len(alt_pop) < 3:
                    alt_pop = np.vstack((alt_pop, self.population))
                selection = self.rng.choice(np.arange(len(alt_pop)), size, replace=False)
                alt_pop = np.stack(alt_pop)
                return np.array(alt_pop)[selection]
        else:
            selection = self.rng.choice(np.arange(len(self.population)), size, replace=False)
            return np.array(self.population)[selection]

    def _crossover_bin(self, target, mutant):
        '''Performs the binomial crossover of DE
        '''
        cross_points = self.rng.rand(self.dimension) < self.crossover_prob
        if not np.any(cross_points):
            cross_points[self.rng.randint(0, self.dimension)] = True
        offspring = np.where(cross_points, mutant, target)
        return offspring

    def _crossover_exp(self, target, mutant):
        '''Performs the exponential crossover of DE
        '''
        n = self.rng.randint(0, self.dimension)
        L = 0
        while ((self.rng.rand() < self.crossover_prob) and L < self.dimension):
            idx = (n+L) % self.dimension
            target[idx] = mutant[idx]
            L = L + 1
        return target

    def crossover(self, target, mutant):
        '''Performs DE crossover
        '''
        if self.crossover_strategy == 'bin':
            offspring = self._crossover_bin(target, mutant)
        elif self.crossover_strategy == 'exp':
            offspring = self._crossover_exp(target, mutant)
        return offspring


opt_class = DE



from typing import Optional, List, Tuple, cast

import numpy as np
import cma
from xbbo.utils.constants import MAXINT
# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trial, Trials
from . import alg_register


@alg_register.register('cma-es')
class CMAES(AbstractOptimizer):
    def __init__(self,
                 space: DenseConfigurationSpace,
                 seed: int = 42,
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)
        # Uniform2Gaussian.__init__(self, )
        if self.space.get_conditions():
            raise NotImplementedError(
                "BORE optimizer currently does not support conditional space!")
        self.dimension = self.space.get_dimensions()
        # self.bounds = self.space.get_bounds()
        self.es = cma.CMAEvolutionStrategy([0.5] * self.dimension, 0.1, inopts={'seed':self.rng.randint(MAXINT),'bounds': [0, 1]})
        # self.hp_num = len(configs)

        self.trials = Trials(dim=self.dimension)
        self.listx = []
        self.listy = []

    def _suggest(self, n_suggestions=1):
        trial_list = []
        for n in range(n_suggestions):
            new_individual = self.es.ask(1)[0]
            # array = self.feature_to_array(new_individual)
            array = new_individual
            config = DenseConfiguration.from_array(self.space,array)
            trial_list.append(
                Trial(config,
                      config_dict=config.get_dictionary(),
                      array=array,
                      origin='cma-es'))

        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial)
            # self.listx.append(self.array_to_feature(trial.array))
            self.listx.append(trial.array)
            self.listy.append(trial.observe_value)
        if len(self.listx) >= self.es.popsize:
            self.es.tell(self.listx, self.listy)
            self.listx = []
            self.listy = []



opt_class = CMAES



import numpy as np
import pickle, os, json
from ConfigSpace.hyperparameters import (CategoricalHyperparameter,
                                         OrdinalHyperparameter, Constant,
                                         UniformFloatHyperparameter,
                                         UniformIntegerHyperparameter)


def dumpOBJ(path, filename, obj):
    with open(os.path.join(path, filename), 'wb') as f:
        pickle.dump(obj, f)

def loadOBJ(filename):
    with open(filename, 'rb') as f:
        obj = pickle.load(f)
    return obj

def dumpJson(path, filename, j):
    with open(os.path.join(path, filename), 'w') as f:
        json.dump(j, f)

def loadJson(filename):
    with open(filename, 'r') as f:
        j = json.load(f)
    return j

def get_types(config_space, instance_features=None):
    """TODO"""
    # Extract types vector for rf from config space and the bounds
    types = [0] * len(config_space.get_hyperparameters())
    bounds = [(np.nan, np.nan)] * len(types)

    for i, param in enumerate(config_space.get_hyperparameters()):
        parents = config_space.get_parents_of(param.name)
        if len(parents) == 0:
            can_be_inactive = False
        else:
            can_be_inactive = True

        if isinstance(param, (CategoricalHyperparameter)):
            n_cats = len(param.choices)
            if can_be_inactive:
                n_cats = len(param.choices) + 1
            types[i] = n_cats
            bounds[i] = (int(n_cats), np.nan)

        elif isinstance(param, (OrdinalHyperparameter)):
            n_cats = len(param.sequence)
            types[i] = 0
            if can_be_inactive:
                bounds[i] = (0, int(n_cats))
            else:
                bounds[i] = (0, int(n_cats) - 1)

        elif isinstance(param, Constant):
            # for constants we simply set types to 0 which makes it a numerical
            # parameter
            if can_be_inactive:
                bounds[i] = (2, np.nan)
                types[i] = 2
            else:
                bounds[i] = (0, np.nan)
                types[i] = 0
            # and we leave the bounds to be 0 for now
        elif isinstance(param, UniformFloatHyperparameter):
            # Are sampled on the unit hypercube thus the bounds
            # are always 0.0, 1.0
            if can_be_inactive:
                bounds[i] = (-1.0, 1.0)
            else:
                bounds[i] = (0, 1.0)
        elif isinstance(param, UniformIntegerHyperparameter):
            if can_be_inactive:
                bounds[i] = (-1.0, 1.0)
            else:
                bounds[i] = (0, 1.0)
        elif not isinstance(
                param,
            (UniformFloatHyperparameter, UniformIntegerHyperparameter,
             OrdinalHyperparameter, CategoricalHyperparameter)):
            raise TypeError("Unknown hyperparameter type %s" % type(param))

    if instance_features is not None:
        types = types + [0] * instance_features.shape[1]

    types = np.array(types, dtype=np.uint)
    bounds = np.array(bounds, dtype=object)
    return types, bounds


'''
Reference: https://github.com/automl/DEHB
'''

from typing import List
import numpy as np

# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.multi_fidelity.hyperband import HB
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trials, Trial
from xbbo.search_algorithm.multi_fidelity.utils.bracket_manager import BasicConfigGenerator
from xbbo.search_algorithm.tpe_optimizer import TPE
from xbbo.utils.constants import MAXINT, Key
from .. import alg_register

class BOHB_CG(BasicConfigGenerator, TPE):
    def __init__(self, cs, budget, max_pop_size, rng, **kwargs) -> None:
        BasicConfigGenerator.__init__(self, cs, budget, max_pop_size, rng, **kwargs)
        TPE.__init__(self, space=cs, **kwargs)
        self.reset(max_pop_size)

alg_marker = 'bohb'

@alg_register.register(alg_marker)
class BOHB(HB):
    name = alg_marker
    def __init__(self,
                 space: DenseConfigurationSpace,
                 budget_bound=[9, 729],
                #  mutation_factor=0.5,
                #  crossover_prob=0.5,
                #  strategy='rand1_bin',
                 eta: int = 3,
                 seed: int = 42,
                 round_limit: int = 1,
                 bracket_limit=np.inf,
                 boundary_fix_type='random',
                 **kwargs):
        HB.__init__(self,
                    space,
                    budget_bound,
                    eta,
                    seed=seed,
                    round_limit=round_limit,
                    bracket_limit=bracket_limit,
                    boundary_fix_type=boundary_fix_type,
                    **kwargs)

        # Uniform2Gaussian.__init__(self,)
        # self._get_max_pop_sizes()
        # self._init_subpop(**kwargs)

    def _init_subpop(self, **kwargs):
        """ List of DE objects corresponding to the budgets (fidelities)
        """
        self.cg = {}
        for i, b in enumerate(self._max_pop_size.keys()):
            self.cg[b] = BOHB_CG(
                self.space,
                seed=self.rng.randint(MAXINT),
                initial_design="random",
                init_budget=0,
                budget=b,
                max_pop_size=self._max_pop_size[b],
                rng=self.rng,
                **kwargs)
            # self.cg[b] = TPE(
            #     self.space,
            #     seed=self.rng.randint(MAXINT),
            #     initial_design="random",
            #     init_budget=0,
            #     **kwargs)
            # self.cg[b].population = [None] * self._max_pop_size[b]
            # self.cg[b].population_fitness = [np.inf] * self._max_pop_size[b]

    def _acquire_candidate(self, bracket, budget):
        """ Generates/chooses a configuration based on the budget and iteration number
        """
        # select a parent/target
                # select a parent/target
        parent_id = self._get_next_idx_for_subpop(budget, bracket)

        if budget != bracket.budgets[0]:
            if bracket.is_new_rung():
                # TODO: check if generalizes to all budget spacings
                lower_budget, num_configs = bracket.get_lower_budget_promotions(
                    budget)
                self._get_promotion_candidate(lower_budget, budget,
                                              num_configs)
            # else: # 每一列中的第一行，随机生成config

        else:
            if bracket.is_new_rung():
                lower_budget, num_configs = bracket.get_lower_budget_promotions(
                    budget)
                self.cg[budget].population_fitness[:] = np.inf
                
            for b in reversed(self.budgets):
                if self.cg[b].kde_models:
                    break 
            # trial = self.cg[b]._suggest()[0]
            trials = self.cg[b]._suggest(1)
            for i in range(len(trials)):
                self.cg[budget].population[parent_id+i] = trials[i].array
                # self.cg[budget].population[parent_id] = trial.array
        # parent_id = self._get_next_idx_for_subpop(budget, bracket)

        target = self.cg[budget].population[parent_id]
        # target = self.fix_boundary(target)
        return target, parent_id

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, True)
            fitness = trial.observe_value
            job_info = trial.info
            # learner_train_time = job_info.get(Key.EVAL_TIME, 0)
            budget = job_info[Key.BUDGET]
            parent_id = job_info['parent_id']
            individual = trial.array  # TODO
            for bracket in self.active_brackets:
                if bracket.bracket_id == job_info['bracket_id']:
                    # registering is IMPORTANT for Bracket Manager to perform SH
                    bracket.register_job(budget)  # may be new row
                    # bracket job complete
                    bracket.complete_job(
                        budget)  # IMPORTANT to perform synchronous SH
            self.cg[budget].population[parent_id] = individual
            self.cg[budget].population_fitness[parent_id] = fitness
            # updating incumbents
            if fitness < self.current_best_fitness:
                self.current_best = individual
                self.current_best_fitness = trial.observe_value
                self.current_best_trial = trial
        # for rung in range(bracket.n_rungs-1, bracket.current_rung, -1):
        #     if self.cg[bracket.budgets[rung]].kde_models:
        #         break
        # else:
            self.cg[budget]._observe([trial])

        self._clean_inactive_brackets()

opt_class = BOHB



# from copy import deepcopy

from .space import DenseConfigurationSpace
import ConfigSpace.hyperparameters as CSH

# from .warp import Warp


def build_space(meta, seed=42):
    """Build Real configspace class.

    Parameters
    ----------
    meta : dict(str, dict)
        Configuration of variables in joint configspace. See API description.
    """
    cs = DenseConfigurationSpace(seed=seed, meta=meta)
    # warp = Warp()
    param_list = sorted(meta.keys())
    # all_warp = {}
    for param_name in param_list:
        # param_name = deepcopy(param_name_)
        param_config = meta[param_name]

        param_type = param_config["type"]

        param_space = param_config.get("warp", 'linear')
        param_range = param_config.get("range", None)
        param_values = param_config.get("values", None)

        # prewarp = 'linear' # None
        if param_type == "cat":  # one-hot
            assert param_range is None
            cs.add_hyperparameter(
                CSH.CategoricalHyperparameter(name=param_name,
                                              choices=param_values))
        elif param_type == "bool":
            assert param_range is None
            assert param_values is None

            cs.add_hyperparameter(
                CSH.CategoricalHyperparameter(name=param_name,
                                              choices=[False, True]))
        elif param_values is not None:
            assert param_type in ("int", "ord", "float"
                                  )  # TODO 有些优化器能直接处理Ordinal？现在直接转为real，warp取整
            assert param_space == 'linear'
            cs.add_hyperparameter(
                CSH.CategoricalHyperparameter(name=param_name,
                                              choices=param_values))
        elif param_type == 'int':
            assert param_values is None

            cs.add_hyperparameter(
                CSH.UniformIntegerHyperparameter(name=param_name,
                                                 lower=param_range[0],
                                                 upper=param_range[1],
                                                 log=param_space == 'log'))
        elif param_type == "float":
            assert param_values is None
            assert param_range is not None
            cs.add_hyperparameter(
                CSH.UniformFloatHyperparameter(name=param_name,
                                               lower=param_range[0],
                                               upper=param_range[1],
                                               log=param_space == 'log'))
        else:
            assert False, "type %s not handled in API" % param_type

    # config_spaces =
    return cs


import numpy as np

from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from . import alg_register
# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.core.trials import Trial, Trials


@alg_register.register('cem')
class CEM(AbstractOptimizer):
    def __init__(self,
                 space: DenseConfigurationSpace,
                 seed: int = 42,
                 llambda=30,
                 elite_ratio=0.3,
                 sample_method: str = 'Gaussian',
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)
        # Uniform2Gaussian.__init__(self, )

        # configs = self.space.get_hyperparameters()
        self.dimension = self.space.get_dimensions()
        self.bounds = self.space.get_bounds()
        if sample_method == 'Gaussian':
            self.sampler = Gaussian_sampler(self.dimension, self.bounds,
                                            self.rng)
        elif sample_method == 'Uniform':
            self.sampler = Uniform_sampler(self.dimension, self.bounds,
                                           self.rng)

        self.buffer_x = []
        self.buffer_y = []
        self.llambda = llambda  #if llambda else 4 + math.floor(3 * math.log(self.dimension))
        self.elite_ratio = elite_ratio
        self.elite_num = max(int(round(self.llambda * self.elite_ratio)), 2)
        self.trials = Trials(dim=self.dimension)

    def _suggest(self, n_suggestions=1):
        trial_list = []
        for n in range(n_suggestions):
            # new_individual = self.feature_to_array(new_individual, )
            new_individual = self.sampler.sample()

            config = DenseConfiguration.from_array(self.space, new_individual)
            trial_list.append(
                Trial(config,
                      config_dict=config.get_dictionary(),
                      array=new_individual,
                      origin='CEM'))

        return trial_list

    def _get_elite(self):
        self.buffer_x = np.asarray(self.buffer_x)
        self.buffer_y = np.asarray(self.buffer_y)
        idx = np.argsort(self.buffer_y)[:self.elite_num]
        return self.buffer_x[idx, :], self.buffer_y[idx]

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)
            self.buffer_x.append(trial.array)
            self.buffer_y.append(trial.observe_value)
        if len(self.buffer_x) < self.llambda:
            return
        elite_x, elite_y = self._get_elite()
        self.sampler.update(elite_x, elite_y)
        self.buffer_x = []
        self.buffer_y = []


class Gaussian_sampler():
    def __init__(self, dim, bounds, rng) -> None:
        self.bounds = bounds
        u = bounds.ub
        l = bounds.lb
        self.mean = (u + l) / 2
        self.std = (u - l) / 6
        self.dim = dim
        self.rng = rng

    def update(self, elite_x, elite_y):
        self.mean = np.mean(elite_x, axis=0)
        self.std = np.std(elite_x, axis=0)

    def sample(self, ):
        new_individual = self.rng.normal(self.mean, self.std + 1e-17)
        new_individual = np.clip(new_individual, self.bounds.lb,
                                 self.bounds.ub)
        return new_individual


class Uniform_sampler():
    def __init__(self, dim, bounds, rng) -> None:
        self.bounds = bounds
        u = bounds.ub
        l = bounds.lb
        self.min = np.copy(l)
        self.max = np.copy(u)
        self.dim = dim
        self.rng = rng

    def update(self, elite_x, elite_y):
        self.min = np.amin(elite_x, axis=0)
        self.max = np.amax(elite_x, axis=0)

    def sample(self, ):
        new_individual = self.rng.uniform(self.min, self.max)
        new_individual = np.clip(new_individual, self.bounds.lb,
                                 self.bounds.ub)
        return new_individual


opt_class = CEM


from collections import defaultdict
import logging
import math
from copy import deepcopy
import warnings
from scipy.stats.qmc import Sobol

import numpy as np
# from xbbo.acquisition_function.acq_optimizer import DesignBoundSearch
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace

from xbbo.core.trials import Trial, Trials
from xbbo.initial_design import ALL_avaliable_design
from xbbo.surrogate.gaussian_process import GPR_sklearn


from xbbo.utils.constants import MAXINT
from . import alg_register

logger = logging.getLogger(__name__)


def latin_hypercube(n_pts, dim):
    """Basic Latin hypercube implementation with center perturbation."""
    X = np.zeros((n_pts, dim))
    centers = (1.0 + 2.0 * np.arange(0.0, n_pts)) / float(2 * n_pts)
    for i in range(dim):  # Shuffle the center locataions for each dimension.
        X[:, i] = centers[np.random.permutation(n_pts)]

    # Add some perturbations within each box
    pert = np.random.uniform(-1.0, 1.0, (n_pts, dim)) / float(2 * n_pts)
    X += pert
    return X


class TuRBO_state():
    def __init__(self,
                 surrogate_model,
                 marker,
                 bounds,
                 rng,
                 dim,
                 n_min_sample,
                 succ_tol=3,
                 length_max=1.6,
                 length_min=0.5**7,
                 length_init=0.8,
                 **kwargs) -> None:
        self.surrogate_model = surrogate_model
        # self.hyper = self.surrogate_models.hypers
        self.dim = dim
        self.marker = marker
        self.rng = rng
        self.bounds = bounds
        self.succ_tol = succ_tol
        self.length_min = length_min
        self.length_max = length_max
        self.length_init = length_init
        self.n_min_sample = n_min_sample
        self.sobol_gen = Sobol(d=self.dim,
                          scramble=True,
                          seed=self.rng.randint(MAXINT))
        self._restart()

    def _restart(self):
        # self.do_optimize = True
        self.length = self.length_init
        self.succ_count = 0
        self.fail_count = 0
        self.center = None
        self.center_value = np.inf

    def _train(self, X, Y):
        # self.surrogate_model.do_optimize = self.do_optimize
        self.surrogate_model.train(X, Y)
        # self.do_optimize = False

    def sample_y(self, X, size=1):
        mean, var = self.surrogate_model.predict_marginalized_over_instances(
            X, 'full_cov')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                sample = self.rng.multivariate_normal(mean.ravel(), var,
                                            size=size).T  # (sample, N)
            except:
                sample = self.rng.multivariate_normal(mean.ravel(), np.diag(np.diag(var)),
                                            size=size).T # TODO
                
        return sample

    def update(self, trial: Trial, trials: Trials, obs_num: int):
        markers = np.array(trials.markers)
        idx = markers == self.marker
        if idx.sum() < self.n_min_sample:
            return
        # self.do_optimize = True
        if (not np.isfinite(self.center_value)
            ) or trial.observe_value < self.center_value - 1e-3 * math.fabs(
                self.center_value):
            self.center = self.to_unit_cube(trial.array)
            self.center_value = trial.observe_value
            logger.info(
                f"{trials.trials_num}) New best @ TR-{self.marker}: {self.center_value:.4}"
            )
            self.succ_count += 1
            self.fail_count = 0
        else:
            self.succ_count = 0
            self.fail_count += obs_num  # NOTE: Add size of the batch for this TR
        if self.succ_count == self.succ_tol:  # Expand trust region
            self.length = min([2.0 * self.length, self.length_max])
            self.succ_count = 0
        elif self.fail_count >= self.fail_tol:  # Shrink trust region (we may have exceeded the fail_tol)
            self.length /= 2.0
            self.fail_count = 0

        # Check if any TR needs to be restarted

        if self.length < self.length_min:  # Restart trust region if converged
            # indicator = [
            #     trial.region == i for trial in self.trials.traj_history
            # ]

            logger.info(
                f"{trials.trials_num}) TR-{self.marker} converged to: : {self.center_value:.4}"
            )

            # Reset length and counters, remove old data from trust region
            self._restart()
            # Remove points from trust region
            markers[idx] = -1
            trials.markers = list(markers)

        X = self.to_unit_cube((trials.get_array())[idx])
        Y = np.array(trials._his_observe_value)[idx]
        self._train(X, Y)
    
    def _get_length_scale(self):
        ks = self.surrogate_model.kernel
        ele_end = 0
        ele_num = None
        for hp in ks.hyperparameters:
            ele_end += hp.n_elements
            if 'length_scale' in hp.name:
                ele_num = hp.n_elements
                break
        assert ele_num is not None
        return np.exp(ks.theta[ele_end-ele_num:ele_end]) # Be careful log theta

    def create_candidates(self, n_candidates):
        length = self.length
        weights = (self._get_length_scale())
        # Create the trust region boundaries
        x_center = self.center  #X[fX.argmin().item(), :][None, :]
        weights = weights / weights.mean(
        )  # This will make the next line more stable
        weights = weights / np.prod(np.power(
            weights, 1.0 / len(weights)))  # We now have weights.prod() = 1
        lb = np.clip(x_center - weights * length / 2.0, 0.0, 1.0)
        ub = np.clip(x_center + weights * length / 2.0, 0.0, 1.0)

        # Draw a Sobolev sequence in [lb, ub]
        
        pert = self.sobol_gen.random(n_candidates)
        pert = lb + (ub - lb) * pert

        # Create a perturbation mask
        prob_perturb = min(20.0 / self.dim, 1.0)
        mask = self.rng.rand(n_candidates, self.dim) <= prob_perturb
        ind = np.where(np.sum(mask, axis=1) == 0)[0]
        mask[ind, self.rng.randint(0, self.dim - 1, size=len(ind))] = 1

        # Create candidate points
        X_cand = x_center.copy() * np.ones((n_candidates, self.dim))
        X_cand[mask] = pert[mask]
        return X_cand

    def to_unit_cube(self, x):
        """Project to [0, 1]^d from hypercube with bounds lb and ub"""

        xx = (x - self.bounds.lb) / (self.bounds.ub - self.bounds.lb)
        return xx

    def from_unit_cube(self, x):
        """Project from [0, 1]^d to hypercube with bounds lb and ub"""
        xx = x * (self.bounds.ub - self.bounds.lb) + self.bounds.lb
        return xx


@alg_register.register('turbo')
class TuRBO(AbstractOptimizer):
    '''
    reference: https://github.com/uber-research/TuRBO/blob/master/turbo/turbo_m.py
    '''
    def __init__(
            self,
            space: DenseConfigurationSpace,
            seed: int = 42,
            surrogate: str = 'gp',
            # acq_func: str = 'mc',
            # acq_opt: str = 'design',
            initial_design: str = 'sobol',
            num_tr=1,
            #  suggest_limit: int = np.inf,
            **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)
        self.dimension = self.space.get_dimensions()
        self.bounds = self.space.get_bounds()
        self.n_min_sample = kwargs.get('n_min_sample', 5)
        self.init_budget = self.n_min_sample  # self.initial_design.init_budget
        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, init_budget=self.init_budget)

        self.initial_design_configs = [[] for _ in range(num_tr)]
        self.trials = Trials(dim=self.dimension)
        self.n_training_steps = kwargs.get("n_training_steps", 50)
        self.max_cholesky_size = kwargs.get("max_cholesky_size", 2000)
        self.dim = self.dimension
        self.n_candidates = 2**int(np.log2(min(100 * self.dim, 5000)))
        self.use_ard = kwargs.get("use_ard", True)
        self.num_tr = num_tr
        self.candidates = []

        if surrogate == 'gp':
            self.turbo_states = [
                TuRBO_state(GPR_sklearn(self.space, rng=self.rng),
                            i,
                            self.bounds,
                            self.rng,
                            self.dim,
                            n_min_sample=self.n_min_sample,
                            **kwargs) for i in range(num_tr)
            ]
            # self.surrogate_models = [
            #     GPR_sklearn(self.space, rng=self.rng) for _ in range(num_tr)
            # ]
        else:
            raise ValueError('surrogate {} not in {}'.format(
                surrogate, ['gp']))

    def _suggest(self, n_suggestions=1):
        markers = np.array(self.trials.markers)

        for m in range(self.num_tr):
            if not hasattr(self.turbo_states[m], 'fail_tol'):
                self.turbo_states[m].fail_tol = np.ceil(
                    np.max([4.0 / n_suggestions, self.dim / n_suggestions]))
            if (markers == m).sum() < self.n_min_sample:
                return self._init_suggest(n_suggestions, m)

        X_cand = np.empty((self.num_tr, self.n_candidates, self.dim))
        y_cand = np.full(
            (self.num_tr, self.n_candidates, n_suggestions), np.inf)
        for m in range(self.num_tr):
            cand = self.turbo_states[m].create_candidates(self.n_candidates)
            cand_y = self.turbo_states[m].sample_y(cand, size=n_suggestions)

            X_cand[m, :, :], y_cand[m, :, :] = cand, cand_y

        X_next = np.empty((n_suggestions, self.dim))
        trial_list = []
        for b in range(n_suggestions):
            marker, j = np.unravel_index(np.argmin(y_cand[:, :, b]),
                                         (self.num_tr, self.n_candidates))
            X_next[b, :] = (X_cand[marker, j, :])
            # X_next[b, :] = deepcopy(X_cand[marker, j, :])
            assert np.isfinite(
                y_cand[marker, j,
                       b])  # Just to make sure we never select nan or inf
            # Make sure we never pick this point again
            y_cand[marker, j, :] = np.inf
            array = self.turbo_states[0].from_unit_cube(X_next[b, :])
            config = DenseConfiguration.from_array(self.space, array)
            trial_list.append(
                Trial(config,
                      config_dict=config.get_dictionary(),
                      array=array,
                      origin='TuRBO-region-{}'.format(marker),
                    #   region=marker,
                      marker=marker))
        return trial_list

    def _observe(self, trial_list):
        update_markers = defaultdict(list)
        for trial in trial_list:
            self.trials.add_a_trial(trial)
            update_markers[trial.marker].append((trial.observe_value, trial))
        for marker in update_markers:
            best_trial = sorted(update_markers[marker],
                                key=lambda x: x[0])[0][1]
            self.turbo_states[marker].update(best_trial, self.trials,
                                             len(trial_list))

        # markers = self.trials.markers
        # for m in update_markers:
        # if (markers == m).sum() <= self.n_min_sample:
        #     continue  # don't train

    def _init_suggest(self, n_suggestions=1, region=0):
        trial_list = []
        for n in range(n_suggestions):
            if not self.initial_design_configs[region]:
                self.initial_design_configs[
                    region] = self.initial_design.select_configurations()
            config = self.initial_design_configs[region].pop(0)
            trial_list.append(
                Trial(
                    configuration=config,
                    config_dict=config.get_dictionary(),
                    array=config.get_array(sparse=False),
                    #   array=config.get_array(sparse=False),
                    origion='turbo-design',
                    # region=region,
                    marker=region))
        return trial_list


opt_class = TuRBO



import importlib
import os
from ConfigSpace import Configuration
import pandas as pd
import numpy as np

from xbbo.core.trials import Trial, Trials

from xbbo.search_algorithm import alg_register
from xbbo.utils.constants import MAXINT, Key
from xbbo.utils.util import dumpJson, dumpOBJ


class BBObenchmark:
    def __init__(self, cfg, seed):
        self.min_budget = None
        self.max_budget = None
        # setup TestProblem
        self.cfg = cfg
        self.problem_name = cfg.TEST_PROBLEM.name
        # self.max_budget = cfg.OPTM.max_budget
        # self.min_budget = cfg.OPTM.min_budget
        self.expdir = cfg.GENERAL.exp_dir
        self.out_dir = os.path.join(self.expdir, self.cfg.OPTM.name)
        if not os.path.exists(self.out_dir):
            os.mkdir(self.out_dir)
        self.problem, self.config_spaces = self._build_problem(
            self.problem_name,
            seed=seed,
            **cfg.TEST_PROBLEM.kwargs)
        self.reset(seed)

    def reset(self, seed):
        self.rng = np.random.RandomState(seed)
        # Setup optimizer
        if self.cfg.OPTM.name in alg_register:
            opt_class = alg_register[self.cfg.OPTM.name]
        else:
            module = importlib.import_module('ext_algs.{}'.format(self.cfg.OPTM.name))
            opt_class = module.opt_class
        
        
        
        self.optimizer_instance = opt_class(
            self.config_spaces,
            seed=self.rng.randint(MAXINT),
            budget_bound=[self.min_budget, self.max_budget],
            objective_function=self._call_obj,
            **dict(self.cfg.OPTM.kwargs))

        self.n_suggestions = self.cfg.OPTM.n_suggestions
        self.n_obj = self.cfg.OPTM.n_obj

        assert self.n_suggestions >= 1, "batch size must be at least 1"
        assert self.n_obj == 1, "Must one objective"

    def _build_problem(self, problem_name: str, seed: int, **kwargs):
        if problem_name == 'countingones':
            from hpolib.benchmarks.synthetic_functions.counting_ones import CountingOnes
            problem = CountingOnes(seed)
            n_continuous = kwargs.get("n_continuous", 4)
            n_categorical = kwargs.get("n_categorical", 4)
            cs = problem.get_configuration_space(n_continuous=n_continuous,
                                                 n_categorical=n_categorical,seed=seed)
            dimensions = len(cs.get_hyperparameters())
            self.min_budget = kwargs.get("min_budget", 576 / dimensions)
            self.max_budget = kwargs.get("max_budget", 93312 / dimensions)
            self.y_star_test = -dimensions
            self._objective_function = problem.objective_function
            self._objective_function_test = problem.objective_function_test
        elif problem_name == 'cc18':
            from hpobench.benchmarks.ml.xgboost_benchmark import XGBoostBenchmark as Benchmark
            task_id = kwargs.get("task_id", 189906)
            # self.min_budget = kwargs.get("min_budget", 0.1)
            # self.max_budget = kwargs.get("max_budget", 1)
            problem = Benchmark(task_id=task_id, rng=seed)
            # Parameter space to be used by DE
            cs = problem.get_configuration_space(seed=seed)
            
            def f_trees(config, budget=None, **kwargs):
                if budget is None:
                    budget = self.max_budget
                fidelity = {
                    "n_estimators": np.round(budget).astype(int),
                    "subsample": 1
                }
                res = problem.objective_function(config,
                                                 fidelity,
                                                 rng=self.rng)
                fitness = res[Key.FUNC_VALUE]
                cost = res[Key.COST]
                res[Key.COST] = res['info']['fidelity']["n_estimators"]
                return res

            def f_dataset(config, budget=None, **kwargs):
                if budget is None:
                    budget = self.max_budget
                fidelity = {"n_estimators": 128, "subsample": budget}
                res = problem.objective_function(config,
                                                 fidelity,
                                                 rng=self.rng)
                fitness = res[Key.FUNC_VALUE]
                cost = res[Key.COST]
                res[Key.COST] = res['info']['fidelity']["subsample"]

                return res

            fidelity_kind = kwargs.get("fidelity_kind", "trees")
            if fidelity_kind == "trees":
                self.min_budget = 50
                self.max_budget = 2000
                self._objective_function = f_trees
            else:
                self.min_budget = 0.1
                self.max_budget = 1.0
                self._objective_function = f_dataset
            self._objective_function_test = lambda config: problem.objective_function(
                config, rng=self.rng)
        elif problem_name == 'svm':
            from hpobench.benchmarks.surrogates.svm_benchmark import SurrogateSVMBenchmark
            problem = SurrogateSVMBenchmark(rng=seed)
            self.min_budget = kwargs.get("min_budget", 1 / 128)
            self.max_budget = kwargs.get("max_budget", 1)
            # Parameter space to be used by DE
            cs = problem.get_configuration_space(seed=seed)

            def f(config, budget=None, **kwargs):
                if budget is not None:
                    fidelity = {"dataset_fraction": budget}
                    res = problem.objective_function(config, fidelity=fidelity)
                else:
                    res = problem.objective_function(config)
                res[Key.COST] = res['info']['fidelity']["dataset_fraction"]

                fitness, cost = res[Key.FUNC_VALUE], res[Key.COST]
                return res

            self._objective_function = f
            self._objective_function_test = problem.objective_function_test
        elif problem_name.startswith('nas_101_cifar10'):
            from hpobench.benchmarks.nas.nasbench_101 import NASCifar10ABenchmark, NASCifar10BBenchmark, NASCifar10CBenchmark
            bench_map = {
                "nas_101_cifar10A": NASCifar10ABenchmark,
                "nas_101_cifar10B": NASCifar10BBenchmark,
                "nas_101_cifar10C": NASCifar10CBenchmark
            }
            self.min_budget = 4
            self.max_budget = 108
            problem = bench_map[problem_name](data_dir='./data',
                                              multi_fidelity=True)

            def f(config, budget=None, **kwargs):
                if budget is None:
                    res = problem.objective_function(config)
                else:
                    res = problem.objective_function(
                        config, fidelity={"dataset_fraction": int(budget)})
                    res[Key.COST] = res['info']['fidelity']["dataset_fraction"]
                return res

            self._objective_function = f
            self._objective_function_test = problem.objective_function_test
            cs = problem.get_configuration_space(seed=seed)
            # y_star_valid = b.y_star_valid
            # y_star_test = b.y_star_test
            # inc_config = None
        elif problem_name.startswith('nas_201'):
            from hpobench.benchmarks.nas.nasbench_201 import NasBench201BaseBenchmark
            dataset = kwargs.get("dataset", "cifar10-valid")
            problem = NasBench201BaseBenchmark(dataset=dataset, rng=seed)
            self.min_budget = 1
            self.max_budget = 200

            def f(config, budget=None, **kwargs):
                if budget is None:
                    res = problem.objective_function(config)
                else:
                    res = problem.objective_function(
                        config, fidelity={"epoch": int(budget)})
                    res[Key.COST] = res['info']['fidelity']["epoch"]
                return res

            self._objective_function = f
            self._objective_function_test = problem.objective_function_test
            cs = problem.get_configuration_space(seed=seed)
            # y_star_valid = b.y_star_valid
            # y_star_test = b.y_star_test
            # inc_config = None
        else:
            raise NotImplementedError

        return problem, cs

    def _call_obj(self, trial, **kwargs):
        if isinstance(trial, Trial):
            config = trial.configuration
        elif isinstance(trial, Configuration):
            config = trial
        else:
            raise NotImplementedError
        budget = kwargs.get(Key.BUDGET)
        r = {}
        if budget is None:
            kwargs[Key.BUDGET] = self.max_budget
        res = self._objective_function(config, **kwargs)
        r.update(kwargs)
        r.update(res)
        res_test = self._objective_function_test(config, **kwargs)
        r[Key.REGRET_TEST] = res_test[Key.FUNC_VALUE]
        r[Key.REGRET_VAL] = res[Key.FUNC_VALUE]
        if Key.COST not in r:
            r[Key.COST] = r.get(Key.BUDGET, kwargs[Key.BUDGET])
        return r

    # def _call_obj_test(self, trial: Trial, **kwargs):
    #     budget = kwargs.get(Key.BUDGET)
    #     r = {}
    #     if budget is None:
    #         kwargs[Key.BUDGET] = self.max_budget
    #     res = self._objective_function_test(trial.configuration, **kwargs)
    #     r.update(kwargs)
    #     r.update(res)
    #     return r

    # def _observe(self, trial_list):
    #     for trial in (trial_list):
    #         info = trial.info.copy()
    #         res = self._call_obj(trial, **info)  # TODO 2
    #         # if self.problem_name in ["countingones"]:
    #         # if hasattr(self, "_call_obj_test"):
    #         #     res_test = self._call_obj_test(trial, **info)  # TODO 2
    #         #     info["regret_test"] = res_test["function_value"]
    #         # info["function_value"] = res["function_value"]
    #         info.update(res)
    #         trial.add_observe_value(observe_value=info['function_value'],
    #                                 obs_info=info)
    #     self.optimizer_instance.observe(trial_list)  # TODO 3

    # def _suggest(self):
    #     return self.optimizer_instance.suggest(self.n_suggestions)  # TODO 1

    def run_one_exp(self):
        self.optimizer_instance.optimize()
        # while not self.optimizer_instance.check_stop():
        #     trial_list = self._suggest()
        #     self._observe(trial_list)
        self.trials = self.optimizer_instance.trials

    def save_to_file(self, run_id):
        trials: Trials = self.trials
        if Key.COST in trials.infos[0]:
            cost_key = Key.COST
        else:
            cost_key = Key.BUDGET
        dumpOBJ(self.out_dir, 'trials_{}.pkl'.format(run_id), trials)
        res = {}
        tmp = np.minimum.accumulate(trials._his_observe_value)
        res[Key.REGRET_VAL] = tmp.tolist()
        if Key.REGRET_TEST in trials.infos[0]:
            res[Key.REGRET_TEST] = np.array(
                [_dict[Key.REGRET_TEST] for _dict in trials.infos])
            res[Key.REGRET_TEST][1:][np.diff(tmp) == 0] = np.nan
            res[Key.REGRET_TEST] = pd.Series(
                res[Key.REGRET_TEST]).fillna(method='ffill').to_list()
        #  = ([_dict['regret_test'] for _dict in trials.infos]).tolist()
        res[Key.COST] = np.cumsum([_dict[cost_key]
                                   for _dict in trials.infos]).tolist()

        dumpJson(self.out_dir, 'res_{}.json'.format(run_id), res)



import logging
import numpy as np
from scipy.optimize import Bounds
from scipy.optimize import minimize, OptimizeResult
from sklearn.ensemble import RandomForestClassifier
from xbbo.core.trials import Trial, Trials
from xbbo.initial_design import ALL_avaliable_design

from . import alg_register
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration

logger = logging.getLogger(__name__)


@alg_register.register('bore')
class BORE(AbstractOptimizer):
    '''
    ref: https://github.com/ltiao/bore
    '''
    def __init__(self,
                 space,
                 seed: int = 42,
                 random_rate: float = 0.1,
                 initial_design: str = 'sobol',
                 suggest_limit: int = np.inf,
                 classify: str = 'rf',
                 **kwargs):

        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='one-hot',
                                   encoding_ord='one-hot',
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)

        # self.multi_start = multi_start(minimizer_fn=minimize)
        self.classifier = Classfify(classify=classify)
        if self.space.get_conditions():
            raise NotImplementedError(
                "BORE optimizer currently does not support conditional space!")

        self.dimension = self.space.get_dimensions()
        bounds = self.space.get_bounds()
        self.bounds = Bounds(bounds.lb, bounds.ub)  #(bounds.lb, bounds.ub)
        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget
        self.hp_num = len(self.space)
        self.initial_design_configs = self.initial_design.select_configurations(
        )

        self.trials = Trials(dim=self.dimension)
        self.random_rate = random_rate
        self.num_starts = kwargs.get("num_starts", 5)
        self.num_samples = kwargs.get("num_samples", 1024)
        self.method = kwargs.get("method", "L-BFGS-B")
        self.options = kwargs.get('options', dict(maxiter=1000, ftol=1e-9))
        self.quantile = kwargs.get("quantile", 0.25)

    def _suggest(self, n_suggestions=1):
        dataset_size = self.trials.trials_num

        # Insufficient training data
        if dataset_size < self.init_budget:
            logger.debug(f"Completed {dataset_size}/{self.init_budget}"
                         " initial runs. Suggesting random candidate...")
            return [
                Trial(configuration=config,
                      config_dict=config.get_dictionary(),
                      array=config.get_array(sparse=False),
                      origin='Random') for config in
                self.initial_design_configs[dataset_size:dataset_size +
                                            n_suggestions]
            ]
        # config_random = []
        # while len(config_random) < n_suggestions:
        #     config = self.space.sample_configuration(1)[0]
        #     if not self.trials.is_contain(config):
        #         config_random.append(config)

        # # epsilon-greedy exploration
        # if self.random_rate is not None and \
        #         self.rng.binomial(p=self.random_rate, n=1):
        #     logger.info("[Glob. maximum: skipped "
        #                 f"(prob={self.random_rate:.2f})] "
        #                 "Suggesting random candidate ...")
        #     return [
        #         Trial(configuration=config,
        #               config_dict=config.get_dictionary(),
        #               array=config.get_array(sparse=False),
        #               origin='Random') for config in config_random
        #     ]
        targets = self.trials.get_history()[0]
        tau = np.quantile(targets, q=self.quantile)
        z = np.less(targets, tau)
        self.classifier.fit(self.trials.get_array(), z)
        # # Create classifier (if retraining from scratch every iteration)
        # self._maybe_create_classifier()
        #
        # # Train classifier
        # self._update_classifier()
        X_init = self.rng.uniform(low=self.bounds.lb,
                                  high=self.bounds.ub,
                                  size=(self.num_samples,
                                        self.dimension))
        f_init = self.classifier.predict(X_init)  # to minimize

        trial_list = []
        for n_ in range(n_suggestions):
            results = []
            if self.num_starts > 0:
                ind = np.argpartition(f_init,
                                      kth=self.num_starts - 1,
                                      axis=None)
                for i in range(self.num_starts):
                    x0 = X_init[ind[i]]
                    result = minimize(self.classifier.predict,
                                      x0=x0,
                                      method=self.method,
                                      jac=False,
                                      bounds=self.bounds,
                                      options=self.options)
                    results.append(result)
                    # # TODO(LT): Make this message a customizable option.
                    # print(f"[Maximum {i+1:02d}: value={result.fun:.3f}] "
                    #         f"success: {result.success}, "
                    #         f"iterations: {result.nit:02d}, "
                    #         f"status: {result.status} ({result.message})")
            else:
                i = np.argmin(f_init, axis=None)
                result = OptimizeResult(x=X_init[i],
                                        fun=f_init[i],
                                        success=True)
                results.append(result)
            best_v = np.inf
            best_config = None
            for res in results:
                if (res.success or res.status == 1) and res.fun < best_v:
                    config = DenseConfiguration.from_array(
                        self.space, res.x)
                    if not self.trials.is_contain(config):
                        best_config = config
            assert best_config is not None
            trial_list.append(
                Trial(configuration=config,
                      config_dict=config.get_dictionary(),
                      array=config.get_array(sparse=False)))

        return trial_list

    def _observe(self, trial_list):
        # for xx, yy in zip(features, y):
        for trial in trial_list:
            self.trials.add_a_trial(trial)

    def _is_unique(self, res, rtol=1e-5, atol=1e-8):
        is_duplicate = any(
            np.allclose(x_prev, res.x, rtol=rtol, atol=atol)
            for x_prev in self.trials.get_array())
        if is_duplicate:
            logger.warn("Duplicate detected! Skipping...")
        return not is_duplicate


class Classfify():
    def __init__(self, classify: str = 'rf'):
        if classify == 'rf':
            self.model = RandomForestClassifier(n_estimators=25)
        else:
            raise NotImplementedError()

    def fit(self, X, z):
        self.model.fit(X, z.ravel())

    def predict(self, x):
        if x.ndim == 1:
            x = x.reshape(1, -1)
        return 1 - self.model.predict_proba(x)[:, 1]


def from_bounds(bounds):

    if isinstance(bounds, Bounds):
        low = bounds.lb
        high = bounds.ub
        dim = len(low)
        assert dim == len(high), "lower and upper bounds sizes do not match!"
    else:
        # assumes `bounds` is a list of tuples
        low, high = zip(*bounds)
        dim = len(bounds)

    return (low, high), dim


opt_class = BORE


'''
Reference: https://github.com/automl/DEHB
'''
import ConfigSpace
import logging
from typing import List
import numpy as np
from sklearn.ensemble import RandomForestClassifier
import statsmodels.api as sm
import scipy.stats as sps
from xbbo.configspace.space import DenseConfiguration
from xbbo.utils.constants import MAXINT, Key

logger = logging.getLogger(__name__)


class SHBracketManager(object):
    """ Synchronous Successive Halving utilities
    """
    def __init__(self, n_configs, budgets, bracket_id=None):
        assert len(n_configs) == len(budgets)
        self.n_configs = n_configs
        self.budgets = budgets
        self.bracket_id = bracket_id
        self.pending_sh_bracket = {}
        self.done_sh_bracket = {}
        self._config_map = {}
        for i, budget in enumerate(budgets):
            # sh_bracket keeps track of jobs/configs that are still to be scheduled/allocatted
            # _sh_bracket keeps track of jobs/configs that have been run and results retrieved for
            # (sh_bracket[i] + _sh_bracket[i]) == n_configs[i] is when no jobs have been scheduled
            #   or all jobs for that budget/rung are over
            # (sh_bracket[i] + _sh_bracket[i]) < n_configs[i] indicates a job has been scheduled
            #   and is queued/running and the bracket needs to be paused till results are retrieved
            self.pending_sh_bracket[budget] = n_configs[
                i]  # each scheduled job does -= 1
            self.done_sh_bracket[budget] = 0  # each retrieved job does +=1
        self.n_rungs = len(budgets)
        self.current_rung = 0

    def is_new_rung(self, ):  # is new row?
        return self.pending_sh_bracket[
            self.get_budget()] == self.current_n_config

    @property
    def current_n_config(self, ):
        return self.n_configs[self.current_rung]

    def get_budget(self, rung=None):
        """ Returns the exact budget that rung is pointing to.

        Returns current rung's budget if no rung is passed.
        """
        if rung is not None:
            return self.budgets[rung]
        return self.budgets[self.current_rung]

    def get_lower_budget_promotions(self, budget):
        """ Returns the immediate lower budget and the number of configs to be promoted from there
        """
        assert budget in self.budgets
        rung = np.where(budget == self.budgets)[0][0]
        prev_rung = np.clip(rung - 1, a_min=0, a_max=self.n_rungs - 1)
        lower_budget = self.budgets[prev_rung]
        num_promote_configs = self.n_configs[rung]
        return lower_budget, num_promote_configs

    def get_next_job_budget(self):
        """ Returns the budget that will be selected if current_rung is incremented by 1
        """
        if self.pending_sh_bracket[self.get_budget()] > 0:
            # the current rung still has unallocated jobs (>0)
            return self.get_budget()
        else:
            print('warnning')
            # the current rung has no more jobs to allocate, increment it
            rung = (self.current_rung + 1) % self.n_rungs
            if self.pending_sh_bracket[self.get_budget(rung)] > 0:
                # the incremented rung has unallocated jobs (>0)
                return self.get_budget(rung)
            else:
                # all jobs for this bracket has been allocated/bracket is complete
                # no more budgets to evaluate and can return None
                pass
            return None

    def register_job(self, budget):
        """ Registers the allocation of a configuration for the budget and updates current rung

        This function must be called when scheduling a job in order to allow the bracket manager
        to continue job and budget allocation without waiting for jobs to finish and return
        results necessarily. This feature can be leveraged to run brackets asynchronously.
        """
        assert budget in self.budgets
        assert self.pending_sh_bracket[budget] > 0
        self.pending_sh_bracket[budget] -= 1
        if not self._is_rung_pending(self.current_rung):
            # increment current rung if no jobs left in the rung
            self.current_rung = (self.current_rung + 1) % self.n_rungs

    def complete_job(self, budget):
        """ Notifies the bracket that a job for a budget has been completed

        This function must be called when a config for a budget has finished evaluation to inform
        the Bracket Manager that no job needs to be waited for and the next rung can begin for the
        synchronous Successive Halving case.
        """
        assert budget in self.budgets
        _max_configs = self.n_configs[list(self.budgets).index(budget)]
        assert self.done_sh_bracket[budget] < _max_configs
        self.done_sh_bracket[budget] += 1

    def _is_rung_waiting(self, rung):
        """ Returns True if at least one job is still pending/running and waits for results
        """
        job_count = self.done_sh_bracket[
            self.budgets[rung]] + self.pending_sh_bracket[self.budgets[rung]]
        if job_count < self.n_configs[rung]:
            return True
        return False

    def _is_rung_pending(self, rung):
        """ Returns True if at least one job pending to be allocatted in the rung
        """
        if self.pending_sh_bracket[self.budgets[rung]] > 0:
            return True
        return False

    def previous_rung_waits(self):
        """ Returns True if none of the rungs < current rung is waiting for results
        """
        for rung in range(self.current_rung):
            if self._is_rung_waiting(rung) and not self._is_rung_pending(rung):
                return True
        return False

    def is_bracket_done(self):
        """ Returns True if all configs in all rungs in the bracket have been allocated
        """
        return ~self.is_pending() and ~self.is_waiting()

    def is_pending(self):
        """ Returns True if any of the rungs/budgets have still a configuration to submit
        """
        return np.any(
            [self._is_rung_pending(i) > 0 for i, _ in enumerate(self.budgets)])

    def is_waiting(self):
        """ Returns True if any of the rungs/budgets have a configuration pending/running
        """
        return np.any(
            [self._is_rung_waiting(i) > 0 for i, _ in enumerate(self.budgets)])

    def __repr__(self):
        cell_width = 9
        cell = "{{:^{}}}".format(cell_width)
        budget_cell = "{{:^{}.2f}}".format(cell_width)
        header = "|{}|{}|{}|{}|".format(cell.format(Key.BUDGET),
                                        cell.format("pending"),
                                        cell.format("waiting"),
                                        cell.format("done"))
        _hline = "-" * len(header)
        table = [header, _hline]
        for i, budget in enumerate(self.budgets):
            pending = self.pending_sh_bracket[budget]
            done = self.done_sh_bracket[budget]
            waiting = np.abs(self.n_configs[i] - pending - done)
            entry = "|{}|{}|{}|{}|".format(budget_cell.format(budget),
                                           cell.format(pending),
                                           cell.format(waiting),
                                           cell.format(done))
            table.append(entry)
        table.append(_hline)
        return "\n".join(table)
class BasicConfigGenerator():
    '''
    every congfig generator only response for one specific budget!
    '''
    def __init__(self, cs, budget, max_pop_size, rng, **kwargs) -> None:
        self.cs = cs
        self.budget = budget
        self.max_pop_size = max_pop_size
        self.rng = rng
        # self.fitness = np.array([np.inf] * self.pop_size)
        # self.reset()
        # self.parent_counter = 0

    def reset(self, pop_size):
        self.inc_score = np.inf
        self.inc_config = None
        self.population = self._init_population(pop_size)
        self.population_fitness = np.array([np.inf] * pop_size)
        # adding attributes to DEHB objects to allow communication across subpopulations
        self.parent_idx = 0
        self.promotion_pop = None
        self.promotion_fitness = None

        # self.history = []

    def _shuffle_pop(self):
        pop_order = np.arange(len(self.population))
        self.rng.shuffle(pop_order)
        self.population = self.population[pop_order]
        self.population_fitness = self.population_fitness[pop_order]
        # self.age = self.age[pop_order]

    def _sort_pop(self):
        pop_order = np.argsort(self.population_fitness)
        self.rng.shuffle(pop_order)
        self.population = self.population[pop_order]
        self.population_fitness = self.population_fitness[pop_order]
        self.age = self.age[pop_order]

    def _init_population(self, pop_size: int) -> List:
        # sample from ConfigSpace s.t. conditional constraints (if any) are maintained
        population = [
            config.get_array(sparse=False)
            for config in self.cs.sample_configuration(size=pop_size)
        ]
        if not isinstance(population, List):
            population = [population]
        # the population is maintained in a list-of-vector form where each ConfigSpace
        return np.array(population)

class ConfigGenerator():
    '''
    every congfig generator only response for one specific budget!
    '''
    def __init__(self, cs, budget, max_pop_size, rng, **kwargs) -> None:
        self.cs = cs
        self.budget = budget
        self.max_pop_size = max_pop_size
        self.rng = rng
        # self.fitness = np.array([np.inf] * self.pop_size)
        # self.reset()
        # self.parent_counter = 0

    def reset(self, pop_size):
        self.inc_score = np.inf
        self.inc_config = None
        self.population = self._init_population(pop_size)
        self.population_fitness = np.array([np.inf] * pop_size)
        # adding attributes to DEHB objects to allow communication across subpopulations
        self.parent_idx = 0
        self.promotion_pop = None
        self.promotion_fitness = None

        # self.history = []

    def _shuffle_pop(self):
        pop_order = np.arange(len(self.population))
        self.rng.shuffle(pop_order)
        self.population = self.population[pop_order]
        self.population_fitness = self.population_fitness[pop_order]
        # self.age = self.age[pop_order]

    def _sort_pop(self):
        pop_order = np.argsort(self.population_fitness)
        self.rng.shuffle(pop_order)
        self.population = self.population[pop_order]
        self.population_fitness = self.population_fitness[pop_order]
        self.age = self.age[pop_order]

    def _init_population(self, pop_size: int) -> List:
        # sample from ConfigSpace s.t. conditional constraints (if any) are maintained
        population = [
            config.get_array(sparse=False)
            for config in self.cs.sample_configuration(size=pop_size)
        ]
        if not isinstance(population, List):
            population = [population]
        # the population is maintained in a list-of-vector form where each ConfigSpace
        return np.array(population)


class DEHB_ConfigGenerator(ConfigGenerator):
    def __init__(self,
                 cs,
                 budget,
                 max_pop_size,
                 rng,
                 mutation_factor=0.5,
                 crossover_prob=0.5,
                 strategy='rand1_bin',
                 **kwargs) -> None:
        super().__init__(cs, budget, max_pop_size, rng, **kwargs)
        self.mutation_factor = mutation_factor
        self.crossover_prob = crossover_prob
        self.dimension = self.cs.get_dimensions()
        self.strategy = strategy
        if self.strategy is not None:
            self.mutation_strategy = self.strategy.split('_')[0]
            self.crossover_strategy = self.strategy.split('_')[1]
        else:
            self.mutation_strategy = self.crossover_strategy = None
        self.reset(max_pop_size)
        # self.population = [None] * self.llambda
        # self.population_fitness = [np.inf] * self.llambda
        # self.current_best = None
        # self.current_best_fitness = np.inf
        self._set_min_pop_size()

    def _set_min_pop_size(self):
        if self.mutation_strategy in ['rand1', 'rand2dir', 'randtobest1']:
            self._min_pop_size = 3
        elif self.mutation_strategy in ['currenttobest1', 'best1']:
            self._min_pop_size = 2
        elif self.mutation_strategy in ['best2']:
            self._min_pop_size = 4
        elif self.mutation_strategy in ['rand2']:
            self._min_pop_size = 5
        else:
            self._min_pop_size = 1

        return self._min_pop_size

    def _mutation_rand1(self, r1, r2, r3):
        '''Performs the 'rand1' type of DE mutation
        '''

        diff = r2 - r3
        mutant = r1 + self.mutation_factor * diff
        return mutant

    def _mutation_rand2(self, r1, r2, r3, r4, r5):
        '''Performs the 'rand2' type of DE mutation
        '''
        diff1 = r2 - r3
        diff2 = r4 - r5
        mutant = r1 + self.mutation_factor * diff1 + self.mutation_factor * diff2
        return mutant

    def _mutation_currenttobest1(self, current, best, r1, r2):
        diff1 = best - current
        diff2 = r1 - r2
        mutant = current + self.mutation_factor * diff1 + self.mutation_factor * diff2
        return mutant

    def _mutation_rand2dir(self, r1, r2, r3):
        diff = r1 - r2 - r3
        mutant = r1 + self.mutation_factor * diff / 2
        return mutant

    # def mutation(self, current=None, best=None, alt_pop=None):
    #     '''Performs DE mutation
    #     '''
    #     if self.mutation_strategy == 'rand1':
    #         r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop)

    #         mutant = self._mutation_rand1(r1, r2, r3)

    #     elif self.mutation_strategy == 'rand2':
    #         r1, r2, r3, r4, r5 = self._sample_population(size=5,
    #                                                      alt_pop=alt_pop)
    #         mutant = self._mutation_rand2(r1, r2, r3, r4, r5)

    #     elif self.mutation_strategy == 'rand2dir':
    #         r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop)

    #         mutant = self._mutation_rand2dir(r1, r2, r3)

    #     elif self.mutation_strategy == 'best1':
    #         r1, r2 = self._sample_population(size=2, alt_pop=alt_pop)

    #         if best is None:
    #             best = self.population[np.argmin(self.population_fitness)]
    #         mutant = self._mutation_rand1(best, r1, r2)

    #     elif self.mutation_strategy == 'best2':
    #         r1, r2, r3, r4 = self._sample_population(size=4, alt_pop=alt_pop)
    #         if best is None:
    #             best = self.population[np.argmin(self.population_fitness)]
    #         mutant = self._mutation_rand2(best, r1, r2, r3, r4)

    #     elif self.mutation_strategy == 'currenttobest1':
    #         r1, r2 = self._sample_population(size=2, alt_pop=alt_pop)
    #         if best is None:
    #             best = self.population[np.argmin(self.population_fitness)]
    #         mutant = self._mutation_currenttobest1(current, best, r1, r2)

    #     elif self.mutation_strategy == 'randtobest1':
    #         r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop)
    #         if best is None:
    #             best = self.population[np.argmin(self.population_fitness)]
    #         mutant = self._mutation_currenttobest1(r1, best, r2, r3)

    #     return mutant

    def mutation(self, current=None, best=None, alt_pop=None):
        '''Performs DE mutation
        '''
        if self.mutation_strategy == 'rand1':
            r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop, target=current)
            mutant = self._mutation_rand1(r1, r2, r3)

        elif self.mutation_strategy == 'rand2':
            r1, r2, r3, r4, r5 = self._sample_population(size=5, alt_pop=alt_pop, target=current)
            mutant = self._mutation_rand2(r1, r2, r3, r4, r5)

        elif self.mutation_strategy == 'rand2dir':
            r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop, target=current)
            mutant = self._mutation_rand2dir(r1, r2, r3)

        elif self.mutation_strategy == 'best1':
            r1, r2 = self._sample_population(size=2, alt_pop=alt_pop, target=current)
            if best is None:
                best = self.population[np.argmin(self.fitness)]
            mutant = self._mutation_rand1(best, r1, r2)

        elif self.mutation_strategy == 'best2':
            r1, r2, r3, r4 = self._sample_population(size=4, alt_pop=alt_pop, target=current)
            if best is None:
                best = self.population[np.argmin(self.fitness)]
            mutant = self._mutation_rand2(best, r1, r2, r3, r4)

        elif self.mutation_strategy == 'currenttobest1':
            r1, r2 = self._sample_population(size=2, alt_pop=alt_pop, target=current)
            if best is None:
                best = self.population[np.argmin(self.fitness)]
            mutant = self._mutation_currenttobest1(current, best, r1, r2)

        elif self.mutation_strategy == 'randtobest1':
            r1, r2, r3 = self._sample_population(size=3, alt_pop=alt_pop, target=current)
            if best is None:
                best = self.population[np.argmin(self.fitness)]
            mutant = self._mutation_currenttobest1(r1, best, r2, r3)

        return mutant

    def _sample_population(self, size=3, alt_pop=None, target=None):
        '''Samples 'size' individuals for mutation step

        If alt_pop is None or a list/array of None, sample from own population
        Else sample from the specified alternate population
        '''
        population = None
        if isinstance(alt_pop, list) or isinstance(alt_pop, np.ndarray):
            idx = [indv is None for indv in alt_pop]  # checks if all individuals are valid
            if any(idx):
                # default to the object's initialized population
                population = self.population
            else:
                # choose the passed population
                population = alt_pop
        else:
            # default to the object's initialized population
            population = self.population

        if target is not None and len(population) > 1:
            # eliminating target from mutation sampling pool
            # the target individual should not be a part of the candidates for mutation
            for i, pop in enumerate(population):
                if all(target == pop):
                    population = np.concatenate((population[:i], population[i + 1:]))
                    break
        if len(population) < self._min_pop_size:
            # compensate if target was part of the population and deleted earlier
            filler = self._min_pop_size - len(population)
            new_pop = self._init_population(pop_size=filler)  # chosen in a uniformly random manner
            population = np.concatenate((population, new_pop))

        selection = self.rng.choice(np.arange(len(population)), size, replace=False)
        return population[selection]
    # def _sample_population(self, size: int = 3, alt_pop=None):
    #     '''Samples 'size' individuals

    #     If alt_pop is None or a list/array of None, sample from own population
    #     Else sample from the specified alternate population (alt_pop)
    #     '''
    #     if isinstance(alt_pop, list) or isinstance(alt_pop, np.ndarray):
    #         idx = [indv is None for indv in alt_pop]
    #         if any(idx):
    #             selection = self.rng.choice(np.arange(len(self.population)),
    #                                         size,
    #                                         replace=False)
    #             return np.array(self.population)[selection]
    #         else:
    #             if len(alt_pop) < 3:
    #                 alt_pop = np.vstack((alt_pop, self.population))
    #             selection = self.rng.choice(np.arange(len(alt_pop)),
    #                                         size,
    #                                         replace=False)
    #             alt_pop = np.stack(alt_pop)
    #             return np.array(alt_pop)[selection]
    #     else:
    #         selection = self.rng.choice(np.arange(len(self.population)),
    #                                     size,
    #                                     replace=False)
    #         return np.array(self.population)[selection]

    def _crossover_bin(self, target, mutant):
        '''Performs the binomial crossover of DE
        '''
        cross_points = self.rng.rand(self.dimension) < self.crossover_prob
        if not np.any(cross_points):
            cross_points[self.rng.randint(0, self.dimension)] = True
        offspring = np.where(cross_points, mutant, target)
        return offspring

    def _crossover_exp(self, target, mutant):
        '''Performs the exponential crossover of DE
        '''
        n = self.rng.randint(0, self.dimension)
        L = 0
        while ((self.rng.rand() < self.crossover_prob) and L < self.dimension):
            idx = (n + L) % self.dimension
            target[idx] = mutant[idx]
            L = L + 1
        return target

    def crossover(self, target, mutant):
        '''Performs DE crossover
        '''
        if self.crossover_strategy == 'bin':
            offspring = self._crossover_bin(target, mutant)
        elif self.crossover_strategy == 'exp':
            offspring = self._crossover_exp(target, mutant)
        return offspring

    def _init_mutant_population(self,
                                pop_size,
                                population,
                                target=None,
                                best=None):
        '''Generates pop_size mutants from the passed population
        '''
        mutants = np.zeros((pop_size, self.dimension))
        for i in range(pop_size):
            mutants[i] = self.mutation(current=target,
                                       best=best,
                                       alt_pop=population)
        return mutants


class RFHB_ConfigGenerator(ConfigGenerator):
    def __init__(self, cs, budget, max_pop_size, rng, eta=3, **kwargs) -> None:
        super().__init__(cs, budget, max_pop_size, rng, **kwargs)
        self.dimension = self.cs.get_dimensions()
        self.reset(max_pop_size)
        self.rf = RandomForestClassifier(n_estimators=25)
        self.reject_rate = kwargs.get("reject_rate", 10)
        self.trained_samples = np.empty((0, self.dimension))
        self.trained_labels = np.empty((0))
        self.fited = False
        self.eta = eta
        self.warmup_scale = kwargs.get("warmup_scale", 3)
        # self.population = [None] * self.llambda
        # self.population_fitness = [np.inf] * self.llambda
        # self.current_best = None
        # self.current_best_fitness = np.inf
        self._set_min_train_size()

    def _set_min_train_size(self):
        self._min_train_size = max(self.warmup_scale * self.dimension, 10)

        return self._min_train_size

    def reset(self, pop_size):
        self.inc_score = np.inf
        self.inc_config = None
        self.parent_idx = 0
        self.population = self._init_population(pop_size)
        self.population_fitness = np.array([np.inf] * pop_size)
        # self.batch_sample_num = int(self.reject_rate * self.pop_size)
    def add_fitting(self, X, y):
        self.trained_samples = np.concatenate([self.trained_samples, X])
        self.trained_labels = np.concatenate([self.trained_labels, y])
        # labels = np.zeros_like(self.trained_labels)

        if len(self.trained_labels) >= self._min_train_size:
            tau = np.quantile(self.trained_labels, q=1 / self.eta)
            labels = self.trained_labels <= tau
            self.rf.fit(self.trained_samples, labels)
            self.fited = True

    def update_new_subpopulation(self, elite_population):
        num_configs = len(elite_population)
        if self.fited:
            batch = self._init_population(
                int(self.reject_rate * num_configs))
            tmp = self.rf.predict_proba(batch)[:, 0]
            tmp[tmp==0] = np.inf
            best_id = np.argsort(tmp)
            self.population[:num_configs] = batch[best_id[:num_configs]]
        else:
            self.population[:num_configs] = elite_population
        self.population_fitness = np.full(len(self.population), np.inf)


class BOHB_ConfigGenerator(ConfigGenerator):
    def __init__(self,
                 cs,
                 budget,
                 max_pop_size,
                 rng,
                 eta=3,
                 gamma=0.15,
                 candidates_num=64,
                 min_bandwidth=1e-3,
                 bandwidth_factor=3,
                 min_points_in_model=None,
                 random_fraction=1 / 3,
                 trials=None,
                 **kwargs) -> None:
        super().__init__(cs, budget, max_pop_size, rng, **kwargs)
        self.dimension = self.cs.get_dimensions()
        self.reset(max_pop_size)
        self.reject_rate = kwargs.get("reject_rate", 10)
        self.trained_samples = np.empty((0, self.dimension))
        self.trained_labels = np.empty((0))
        self.fited = False
        self.eta = eta
        self.min_bandwidth = min_bandwidth
        self.bw_factor = bandwidth_factor
        self.warmup_scale = kwargs.get("warmup_scale", 3)
        # self.population = [None] * self.llambda
        # self.population_fitness = [np.inf] * self.llambda
        # self.current_best = None
        # self.current_best_fitness = np.inf
        self.gamma = gamma
        self.candidates_num = candidates_num
        self.min_points_in_model = min_points_in_model

        hps = self.cs.get_hyperparameters()

        if min_points_in_model is None:
            self.min_points_in_model = len(hps) + 1

        if self.min_points_in_model < len(hps) + 1:
            self.min_points_in_model = len(hps) + 1

        self.random_fraction = random_fraction
        self.kde_models = dict()

        self.kde_vartypes = ""
        self.vartypes = []

        self.population = np.empty((0, self.dimension))
        self.trials = trials
        self.population_fitness = np.empty(0)

    def add_obs(self, population, population_fitness):
        self.population = np.concatenate([self.population, population], axis=0)
        self.population_fitness = np.concatenate(
            [self.population_fitness, population_fitness])

    def reset(self, pop_size):
        self.inc_score = np.inf
        self.inc_config = None
        self.parent_idx = 0
        # self.population = self._init_population(pop_size)
        # self.population_fitness = np.array([np.inf] * pop_size)
        # self.batch_sample_num = int(self.reject_rate * self.pop_size)
    def _sample_nonduplicate_config(self, num_configs=1):
        configs = list()
        sample_cnt = 0
        max_sample_cnt = 1000
        while len(configs) < num_configs:
            config = self.cs.sample_configuration()[0]
            sample_cnt += 1
            if (self.trials is None or not self.trials.is_contain(config)
                ) and config not in configs:
                configs.append(config)
                sample_cnt = 0
                continue
            if sample_cnt >= max_sample_cnt:
                logger.warning(
                    'Cannot sample non duplicate configuration after %d iterations.'
                    % max_sample_cnt)
                configs.append(config)
                sample_cnt = 0
        return configs

    def get_config(self):
        self._fit_kde_models()
        if len(self.kde_models.keys()
               ) == 0 or self.rng.rand() < self.random_fraction:
            config = self._sample_nonduplicate_config()[0]
        else:
            best = np.inf
            best_vector = None
            l = self.kde_models['good'].pdf
            g = self.kde_models['bad'].pdf

            minimize_me = lambda x: max(1e-32, g(x)) / max(l(x), 1e-32)

            kde_good = self.kde_models['good']
            kde_bad = self.kde_models['bad']

            for i in range(self.candidates_num):
                idx = self.rng.randint(0, len(kde_good.data))
                datum = kde_good.data[idx]
                vector = []

                for m, bw, t in zip(datum, kde_good.bw, self.vartypes):

                    bw = max(bw, self.min_bandwidth)
                    if t == 0:
                        bw = self.bw_factor * bw
                        try:
                            vector.append(
                                sps.truncnorm.rvs(-m / bw, (1 - m) / bw,
                                                  loc=m,
                                                  scale=bw))
                        except:
                            logger.warning(
                                "Truncated Normal failed for:\ndatum=%s\nbandwidth=%s\nfor entry with value %s"
                                % (datum, kde_good.bw, m))
                            logger.warning("data in the KDE:\n%s" %
                                           kde_good.data)
                    else:

                        if self.rng.rand() < (1 - bw):
                            vector.append(int(m))
                        else:
                            vector.append(self.rng.randint(t))
                val = minimize_me(vector)

                if not np.isfinite(val):
                    logger.warning('sampled vector: %s has EI value %s' %
                                   (vector, val))
                    logger.warning("data in the KDEs:\n%s\n%s" %
                                   (kde_good.data, kde_bad.data))
                    logger.warning("bandwidth of the KDEs:\n%s\n%s" %
                                   (kde_good.bw, kde_bad.bw))
                    logger.warning("l(x) = %s" % (l(vector)))
                    logger.warning("g(x) = %s" % (g(vector)))

                    # right now, this happens because a KDE does not contain all values for a categorical parameter
                    # this cannot be fixed with the statsmodels KDE, so for now, we are just going to evaluate this one
                    # if the good_kde has a finite value, i.e. there is no config with that value in the bad kde, so it shouldn't be terrible.
                    if np.isfinite(l(vector)):
                        best_vector = vector
                        break

                if val < best:
                    best = val
                    best_vector = vector

            if best_vector is None:
                logger.debug(
                    "Sampling based optimization with %i samples failed -> using random configuration"
                    % self.candidates_num)
                config = self._sample_nonduplicate_config()[0]
            else:
                logger.debug('best_vector: {}, {}, {}, {}'.format(
                    best_vector, best, l(best_vector), g(best_vector)))
                for i, hp_value in enumerate(best_vector):
                    if isinstance(
                            self.cs.get_hyperparameter(
                                self.cs.get_hyperparameter_by_idx(i)),
                            ConfigSpace.hyperparameters.
                            CategoricalHyperparameter):
                        best_vector[i] = int(np.rint(best_vector[i]))

                config = DenseConfiguration.from_array(self.cs,
                                                       np.asarray(best_vector))
        array = config.get_array(sparse=False)

        return array

    def _fit_kde_models(self, ):
        train_configs = self.population
        n_good = max(self.min_points_in_model,
                     int(self.gamma * len(self.population_fitness)) // 100)
        # n_bad = min(max(self.min_points_in_model, ((100-self.top_n_percent)*train_configs.shape[0])//100), 10)
        n_bad = max(self.min_points_in_model,
                    int((1 - self.gamma) * len(self.population_fitness)))

        # Refit KDE for the current budget
        idx = np.argsort(self.population_fitness)

        train_data_good = self.impute_conditional_data(
            train_configs[idx[:n_good]])
        train_data_bad = self.impute_conditional_data(
            train_configs[idx[n_good:n_good + n_bad]])

        if train_data_good.shape[0] <= train_data_good.shape[1]:
            return
        if train_data_bad.shape[0] <= train_data_bad.shape[1]:
            return

        bw_estimation = 'normal_reference'
        np.random.seed(self.rng.randint(MAXINT))
        bad_kde = sm.nonparametric.KDEMultivariate(data=train_data_bad,
                                                   var_type=self.kde_vartypes,
                                                   bw=bw_estimation)
        good_kde = sm.nonparametric.KDEMultivariate(data=train_data_good,
                                                    var_type=self.kde_vartypes,
                                                    bw=bw_estimation)

        bad_kde.bw = np.clip(bad_kde.bw, self.min_bandwidth, None)
        good_kde.bw = np.clip(good_kde.bw, self.min_bandwidth, None)

        self.kde_models = {'good': good_kde, 'bad': bad_kde}

    def impute_conditional_data(self, array):

        return_array = np.empty_like(array)

        for i in range(array.shape[0]):
            datum = np.copy(array[i])
            nan_indices = np.argwhere(np.isnan(datum)).ravel()

            while np.any(nan_indices):
                nan_idx = nan_indices[0]
                valid_indices = np.argwhere(np.isfinite(
                    array[:, nan_idx])).ravel()

                if len(valid_indices) > 0:
                    # pick one of them at random and overwrite all NaN values
                    row_idx = self.rng.choice(valid_indices)
                    datum[nan_indices] = array[row_idx, nan_indices]

                else:
                    # no good point in the data has this value activated, so fill it with a valid but random value
                    t = self.vartypes[nan_idx]
                    if t == 0:
                        datum[nan_idx] = self.rng.rand()
                    else:
                        datum[nan_idx] = self.rng.randint(t)

                nan_indices = np.argwhere(np.isnan(datum)).ravel()
            return_array[i, :] = datum
        return return_array

class RFDEHB_ConfigGenerator(DEHB_ConfigGenerator):
    def __init__(self, cs, budget, max_pop_size, rng, eta=3, **kwargs) -> None:
        super().__init__(cs, budget, max_pop_size, rng, **kwargs)
        self.dimension = self.cs.get_dimensions()
        self.rf = RandomForestClassifier(n_estimators=25)
        self.reject_rate = kwargs.get("reject_rate", 10)
        self.trained_samples = np.empty((0, self.dimension))
        self.trained_labels = np.empty((0))
        self.fited = False
        self.eta = eta
        self.warmup_scale = kwargs.get("warmup_scale", 3)
        # self.population = [None] * self.llambda
        # self.population_fitness = [np.inf] * self.llambda
        # self.current_best = None
        # self.current_best_fitness = np.inf
        self._set_min_train_size()

    def _set_min_train_size(self):
        self._min_train_size = max(self.warmup_scale * self.dimension, 10)

        return self._min_train_size

        # self.batch_sample_num = int(self.reject_rate * self.pop_size)
    def add_fitting(self, X, y):
        self.trained_samples = np.concatenate([self.trained_samples, X])
        self.trained_labels = np.concatenate([self.trained_labels, y])
        # labels = np.zeros_like(self.trained_labels)

        if len(self.trained_labels) >= self._min_train_size:
            tau = np.quantile(self.trained_labels, q=1 / self.eta)
            labels = self.trained_labels <= tau
            self.rf.fit(self.trained_samples, labels)
            self.fited = True

    # def update_new_subpopulation(self, elite_population):
    #     num_configs = len(elite_population)
    #     if self.fited:
    #         batch = self._init_population(
    #             int(self.reject_rate * num_configs))
    #         best_id = np.argsort(self.rf.predict_proba(batch)[:, 0])
    #         self.population[:num_configs] = batch[best_id[:num_configs]]
    #     else:
    #         self.population[:num_configs] = elite_population
    #     self.population_fitness = np.full(len(self.population), np.inf)
    
    def batch_select_mut_cross_child(self,current=None, best=None, alt_pop=None, num_configs=0):
        if self.fited:
            batch_size = int(self.reject_rate * num_configs)
            best_id = None
        else:
            batch_size = 1
            best_id = 0
        individuals = []
        for i in range(batch_size):
            mutant = self.mutation(current,
                                          best=best,
                                          alt_pop=alt_pop)
            # perform crossover with selected parent
            individuals.append(self.crossover(target=current, mutant=mutant))
        if best_id is None:
            # tmp = self.rf.predict_proba(np.array(individuals))[:, 0]
            # idxs = np.argwhere(tmp>0)
            tmp = self.rf.predict_proba(np.array(individuals))[:, 0]
            tmp[tmp==0] = np.inf
            best_id = np.argmin(tmp).item()
            # if len(idxs) == 0:
            #     best_id = 0
            # else:
            #     best_id = np.argmin(tmp[idxs])
            #     best_id = idxs[best_id].item()
        return individuals[best_id]


import typing
import numpy as np

from xbbo.utils.constants import VERY_SMALL_NUMBER
from xbbo.surrogate.gaussian_process import GPR_sklearn

class BaseModel(GPR_sklearn):
    def __init__(
        self,
        cs,
        rng=np.random.RandomState(0),
        n_opt_restarts: int = 10,
        instance_features: typing.Optional[np.ndarray] = None,
        pca_components: typing.Optional[int] = None,
        **kwargs
    ):
        super().__init__(cs, rng, n_opt_restarts,instance_features=instance_features,
            pca_components=pca_components,**kwargs)
        self.cached = {}
        self.cached_normalize = {}

    def _predict_normalize(self, X_test, cov_return_type: typing.Optional[str] = 'diagonal_cov'):
        assert self.is_fited
        # key = hash(X_test.data.tobytes()+bytes(cov_return_type if cov_return_type else '', 'utf-8'))
        # if key in self.cached_normalize:
        #     return self.cached_normalize[key]
        
        X_test = self._impute_inactive(X_test)
        if cov_return_type is None:
            mu = self.gp.predict(X_test)
            var = None

            # if self.normalize_y:
            #     mu = self._untransform_y(mu)

        else:
            predict_kwargs = {'return_cov': False, 'return_std': True}
            if cov_return_type == 'full_cov':
                predict_kwargs = {'return_cov': True, 'return_std': False}

            mu, var = self.gp.predict(X_test, **predict_kwargs)

            if cov_return_type != 'full_cov':
                var = var**2  # since we get standard deviation for faster computation

            # Clip negative variances and set them to the smallest
            # positive float value
            var = np.clip(var, VERY_SMALL_NUMBER, np.inf)

            # if self.normalize_y:
            #     mu, var = self._untransform_y(mu, var)

            if cov_return_type == 'diagonal_std':
                var = np.sqrt(
                    var)  # converting variance to std deviation if specified
        # self.cached_normalize[key] = (mu, var)
        
        return mu, var
    
    def _predict(self, X_test, cov_return_type: typing.Optional[str] = 'diagonal_cov'):
        assert self.is_fited
        # key = hash(X_test.data.tobytes()+bytes(cov_return_type if cov_return_type else '', 'utf-8'))
        # if key in self.cached:
        #     return self.cached[key]
        
        X_test = self._impute_inactive(X_test)
        if cov_return_type is None:
            mu = self.gp.predict(X_test)
            var = None

            if self.normalize_y:
                mu = self._untransform_y(mu)

        else:
            predict_kwargs = {'return_cov': False, 'return_std': True}
            if cov_return_type == 'full_cov':
                predict_kwargs = {'return_cov': True, 'return_std': False}

            mu, var = self.gp.predict(X_test, **predict_kwargs)

            if cov_return_type != 'full_cov':
                var = var**2  # since we get standard deviation for faster computation

            # Clip negative variances and set them to the smallest
            # positive float value
            var = np.clip(var, VERY_SMALL_NUMBER, np.inf)

            if self.normalize_y:
                mu, var = self._untransform_y(mu, var)

            if cov_return_type == 'diagonal_std':
                var = np.sqrt(
                    var)  # converting variance to std deviation if specified
        # self.cached[key] = (mu, var)
        
        return mu, var   

    # def _cached_predict(self, X_test, cov_return_type: typing.Optional[str] = 'diagonal_cov'):
    #     key = hash(X_test.data.tobytes()+bytes(cov_return_type if cov_return_type else '', 'utf-8'))
    #     if key in self.cached:
    #         return self.cached[key]
    #     if key not in self.cached:
    #         self.cached[key] = self._predict(X_test, cov_return_type)
    #     return self.cached[key]
    




import logging
from typing import Iterable, Iterator, List, Optional, Tuple, Union
import numpy as np
import time
import scipy
from scipy.stats.qmc import Sobol


from xbbo.acquisition_function.base import AbstractAcquisitionFunction, AcquisitionFunctionMaximizer

from xbbo.configspace.space import DenseConfigurationSpace, DenseConfiguration, get_one_exchange_neighbourhood
from xbbo.core.trials import Trials
from xbbo.utils.constants import MAXINT
from xbbo.utils.util import get_types

logger = logging.getLogger(__name__)


class RandomSearch(AcquisitionFunctionMaximizer):
    """Get candidate solutions via random sampling of configurations.

    Parameters
    ----------
    acquisition_function : ~xbbo.optimizer.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.configspace.DenseConfigurationSpace

    rng : np.random.RandomState or int, optional
    """
    def _maximize(self,
                  trials: Trials,
                  num_points: int,
                  _sorted: bool = False,
                  **kwargs) -> List[Tuple[float, DenseConfiguration]]:
        """Randomly sampled configurations

        Parameters
        ----------
        trials: ~xbbo.trials.trials.trials
            trials object
        stats: ~xbbo.stats.stats.Stats
            current stats object
        num_points: int
            number of points to be sampled
        _sorted: bool
            whether random configurations are sorted according to acquisition function

        Returns
        -------
        iterable
            An iterable consistng of
            tuple(acqusition_value, :class:`xbbo.configspace.DenseConfiguration`).
        """
        if num_points > 1:
            rand_configs = self.config_space.sample_configuration(
                size=num_points)
        else:
            rand_configs = [self.config_space.sample_configuration(size=1)]
        if _sorted:
            for i in range(len(rand_configs)):
                rand_configs[i].origin = 'Random Search (sorted)'
            return self._sort_configs_by_acq_value(rand_configs)
        else:
            for i in range(len(rand_configs)):
                rand_configs[i].origin = 'Random Search'
            return [(0, rand_configs[i]) for i in range(len(rand_configs))]


class LocalSearch(AcquisitionFunctionMaximizer):
    """Implementation of xbbo's local search.

    Parameters
    ----------
    acquisition_function : ~xbbo.acquisition_function.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.config_space.DenseConfigurationSpace

    rng : np.random.RandomState or int, optional

    max_steps: int
        Maximum number of iterations that the local search will perform

    n_steps_plateau_walk: int
        number of steps during a plateau walk before local search terminates

    """
    def __init__(
        self,
        acquisition_function: AbstractAcquisitionFunction,
        config_space: DenseConfigurationSpace,
        rng: np.random.RandomState = np.random.RandomState(42),
        max_steps: Optional[int] = None,
        n_steps_plateau_walk: int = 10,
    ):
        super().__init__(acquisition_function, config_space, rng)
        self.max_steps = max_steps
        self.n_steps_plateau_walk = n_steps_plateau_walk

    def _maximize(self, trials: Trials, num_points: int,
                  **kwargs) -> List[Tuple[float, DenseConfiguration]]:
        """Starts a local search from the given startpoint and quits
        if either the max number of steps is reached or no neighbor
        with an higher improvement was found.

        Parameters
        ----------
        trials: ~xbbo.utils.history_container.Trials
            trials object
        stats: ~xbbo.stats.stats.Stats
            current stats object
        num_points: int
            number of points to be sampled
        ***kwargs:
            Additional parameters that will be passed to the
            acquisition function

        Returns
        -------
        incumbent: np.array(1, D)
            The best found DenseConfiguration
        acq_val_incumbent: np.array(1,1)
            The acquisition value of the incumbent

        """

        init_points = self._get_initial_points(num_points, trials)

        acq_configs = []
        # Start N local search from different random start points
        for start_point in init_points:
            acq_val, DenseConfiguration = self._one_iter(start_point, **kwargs)

            DenseConfiguration.origin = "Local Search"
            acq_configs.append((acq_val, DenseConfiguration))

        # shuffle for random tie-break
        self.rng.shuffle(acq_configs)

        # sort according to acq value
        acq_configs.sort(reverse=True, key=lambda x: x[0])

        return acq_configs

    def _get_initial_points(self, num_points, trials):

        if trials.is_empty():
            init_points = self.config_space.sample_configuration(
                size=num_points)
        else:
            # initiate local search with best configurations from previous runs
            configs_previous_runs = trials.get_all_configs()
            configs_previous_runs_sorted = self._sort_configs_by_acq_value(
                configs_previous_runs)
            num_configs_local_search = int(
                min(len(configs_previous_runs_sorted), num_points))
            init_points = list(
                map(lambda x: x[1],
                    configs_previous_runs_sorted[:num_configs_local_search]))

        return init_points

    def _one_iter(self, start_point: DenseConfiguration,
                  **kwargs) -> Tuple[float, DenseConfiguration]:

        incumbent = start_point
        # Compute the acquisition value of the incumbent
        acq_val_incumbent = self.acquisition_function([incumbent], **kwargs)[0]

        local_search_steps = 0
        neighbors_looked_at = 0
        time_n = []
        while True:

            local_search_steps += 1
            if local_search_steps % 1000 == 0:
                logger.warning(
                    "Local search took already %d iterations. Is it maybe "
                    "stuck in a infinite loop?", local_search_steps)

            # Get neighborhood of the current incumbent
            # by randomly drawing configurations
            changed_inc = False

            # Get one exchange neighborhood returns an iterator (in contrast of
            # the previously returned list).
            all_neighbors = get_one_exchange_neighbourhood(
                incumbent, seed=self.rng.randint(MAXINT))

            for neighbor in all_neighbors:
                s_time = time.time()
                acq_val = self.acquisition_function([neighbor], **kwargs)
                neighbors_looked_at += 1
                time_n.append(time.time() - s_time)

                if acq_val > acq_val_incumbent:
                    logger.debug("Switch to one of the neighbors")
                    incumbent = neighbor
                    acq_val_incumbent = acq_val
                    changed_inc = True
                    break

            if (not changed_inc) or \
                    (self.max_steps is not None and
                     local_search_steps == self.max_steps):
                logger.debug(
                    "Local search took %d steps and looked at %d "
                    "configurations. Computing the acquisition "
                    "value for one DenseConfiguration took %f seconds"
                    " on average.", local_search_steps, neighbors_looked_at,
                    np.mean(time_n))
                break

        return acq_val_incumbent, incumbent


class ScipyGlobalOptimizer(AcquisitionFunctionMaximizer):
    """
    Wraps scipy global optimizer. Only on continuous dims.

    Parameters
    ----------
    acquisition_function : ~xbbo.acquisition_function.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.config_space.DenseConfigurationSpace

    rng : np.random.RandomState or int, optional
    """
    def __init__(self,
                 acquisition_function: AbstractAcquisitionFunction,
                 config_space: DenseConfigurationSpace,
                 rng: np.random.RandomState = np.random.RandomState(42)):
        super().__init__(acquisition_function, config_space, rng)

        types, bounds = get_types(self.config_space)
        assert all(types == 0)
        self.bounds = bounds

    def maximize(self,
                 trials: Trials,
                 initial_config=None,
                 drop_self_duplicate: bool = False,
                 **kwargs) -> List[Tuple[float, DenseConfiguration]]:
        def negative_acquisition(x):
            # shape of x = (d,)
            return -self.acquisition_function(x,
                                              convert=False)[0]  # shape=(1,)

        acq_configs = []
        result = scipy.optimize.differential_evolution(
            func=negative_acquisition, bounds=self.bounds)
        if not result.success:
            logger.debug(
                'Scipy differential evolution optimizer failed. Info:\n%s' %
                (result, ))
        try:
            config = DenseConfiguration(self.config_space, vector=result.x)
            acq = self.acquisition_function(result.x, convert=False)
            acq_configs.append((acq, config))
        except Exception:
            pass

        if not acq_configs:  # empty
            logger.warning(
                'Scipy differential evolution optimizer failed. Return empty config list. Info:\n%s'
                % (result, ))
        configs = [config for _, config in acq_configs]
        return self.unique(configs=configs) if drop_self_duplicate else configs

    def _maximize(self, trials: Trials, num_points: int,
                  **kwargs) -> Iterable[Tuple[float, DenseConfiguration]]:
        raise NotImplementedError()


class RandomScipyOptimizer(AcquisitionFunctionMaximizer):
    """
    Use scipy.optimize with start points chosen by random search. Only on continuous dims.

    Parameters
    ----------
    acquisition_function : ~xbbo.acquisition_function.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.config_space.DenseConfigurationSpace

    rng : np.random.RandomState or int, optional
    """
    def __init__(
        self,
        acquisition_function: AbstractAcquisitionFunction,
        config_space: DenseConfigurationSpace,
        rng: Union[bool, np.random.RandomState] = None,
    ):
        super().__init__(acquisition_function, config_space, rng)

        self.random_search = InterleavedLocalAndRandomSearch(
            acquisition_function=acquisition_function,
            config_space=config_space,
            rng=rng)
        self.scipy_optimizer = ScipyOptimizer(
            acquisition_function=acquisition_function,
            config_space=config_space,
            rng=rng)

    def maximize(self,
                 trials: Trials,
                 num_points: int,
                 drop_self_duplicate: bool = False,
                 num_trials=10,
                 **kwargs) -> List[Tuple[float, DenseConfiguration]]:
        acq_configs = []

        initial_configs = self.random_search.maximize(trials, num_points,
                                                      **kwargs).challengers
        initial_acqs = self.acquisition_function(initial_configs)
        acq_configs.extend(zip(initial_acqs, initial_configs))

        success_count = 0
        for config in initial_configs[:num_trials]:
            scipy_configs = self.scipy_optimizer.maximize(
                trials, initial_config=config).challengers
            if not scipy_configs:  # empty
                continue
            scipy_acqs = self.acquisition_function(scipy_configs)
            acq_configs.extend(zip(scipy_acqs, scipy_configs))
            success_count += 1
        if success_count == 0:
            logger.warning(
                'None of Scipy optimizations are successful in RandomScipyOptimizer.'
            )

        # shuffle for random tie-break
        self.rng.shuffle(acq_configs)

        # sort according to acq value
        acq_configs.sort(reverse=True, key=lambda x: x[0])

        configs = [_[1] for _ in acq_configs]

        return self.unique(configs=configs) if drop_self_duplicate else configs

    def _maximize(self, trials: Trials, num_points: int,
                  **kwargs) -> Iterable[Tuple[float, DenseConfiguration]]:
        raise NotImplementedError()


class ScipyOptimizer(AcquisitionFunctionMaximizer):
    """
    Wraps scipy optimizer. Only on continuous dims.

    Parameters
    ----------
    acquisition_function : ~xbbo.acquisition_function.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.config_space.DenseConfigurationSpace

    rng : np.random.RandomState or int, optional
    """
    def __init__(
        self,
        acquisition_function: AbstractAcquisitionFunction,
        config_space: DenseConfigurationSpace,
        rng: Union[bool, np.random.RandomState] = None,
    ):
        super().__init__(acquisition_function, config_space, rng)

        types, bounds = get_types(
            self.config_space)  # todo: support constant hp in scipy optimizer
        assert all(
            types == 0
        ), 'Scipy optimizer (L-BFGS-B) only supports Integer and Float parameters.'
        self.bounds = bounds

        options = dict(disp=False, maxiter=1000)
        self.scipy_config = dict(tol=None, method='L-BFGS-B', options=options)

    def maximize(self,
                 trials: Trials,
                 initial_config=None,
                 drop_self_duplicate: bool = False,
                 **kwargs) -> List[Tuple[float, DenseConfiguration]]:
        def negative_acquisition(x):
            # shape of x = (d,)
            x = np.clip(x, 0.0, 1.0)  # fix numerical problem in L-BFGS-B
            return -self.acquisition_function(x,
                                              convert=False)[0]  # shape=(1,)

        if initial_config is None:
            initial_config = self.config_space.sample_configuration()
        init_point = initial_config.get_array(sparse=False)

        acq_configs = []
        result = scipy.optimize.minimize(fun=negative_acquisition,
                                         x0=init_point,
                                         bounds=self.bounds,
                                         **self.scipy_config)
        # if result.success:
        #     acq_configs.append((result.fun, DenseConfiguration(self.config_space, vector=result.x)))
        if not result.success:
            logger.debug('Scipy optimizer failed. Info:\n%s' % (result, ))
        try:
            x = np.clip(result.x, 0.0,
                        1.0)  # fix numerical problem in L-BFGS-B
            config = DenseConfiguration(self.config_space, vector=x)
            acq = self.acquisition_function(x, convert=False)
            acq_configs.append((acq, config))
        except Exception:
            pass

        if not acq_configs:  # empty
            logger.warning(
                'Scipy optimizer failed. Return empty config list. Info:\n%s' %
                (result, ))

        configs = [config for _, config in acq_configs]
        return self.unique(configs=configs) if drop_self_duplicate else configs

    def _maximize(self, trials: Trials, num_points: int,
                  **kwargs) -> Iterable[Tuple[float, DenseConfiguration]]:
        raise NotImplementedError()


class InterleavedLocalAndRandomSearch(AcquisitionFunctionMaximizer):
    """Implements xbbo's default acquisition function optimization.

    This acq_maximizer performs local search from the previous best points
    according, to the acquisition function, uses the acquisition function to
    sort randomly sampled configurations and interleaves unsorted, randomly
    sampled configurations in between.

    Parameters
    ----------
    acquisition_function : ~xbbo.acquisition_function.acquisition.AbstractAcquisitionFunction

    config_space : ~xbbo.config_space.DenseConfigurationSpace

    rng : np.random.RandomState or int, optional

    max_steps: int
        [LocalSearch] Maximum number of steps that the local search will perform

    n_steps_plateau_walk: int
        [LocalSearch] number of steps during a plateau walk before local search terminates

    n_sls_iterations: int
        [Local Search] number of local search iterations

    """
    def __init__(
        self,
        acquisition_function: AbstractAcquisitionFunction,
        config_space: DenseConfigurationSpace,
        rng: Union[bool, np.random.RandomState] = None,
        max_steps: Optional[int] = None,
        n_steps_plateau_walk: int = 10,
        n_sls_iterations: int = 10,
    ):
        super().__init__(acquisition_function, config_space, rng)
        self.random_search = RandomSearch(
            acquisition_function=acquisition_function,
            config_space=config_space,
            rng=rng)
        self.local_search = LocalSearch(
            acquisition_function=acquisition_function,
            config_space=config_space,
            rng=rng,
            max_steps=max_steps,
            n_steps_plateau_walk=n_steps_plateau_walk)
        self.n_sls_iterations = n_sls_iterations

        # =======================================================================
        # self.local_search = DiffOpt(
        #     acquisition_function=acquisition_function,
        #     config_space=config_space,
        #     rng=rng
        # )
        # =======================================================================

    def maximize(self,
                 trials: Trials,
                 num_points: int,
                 drop_self_duplicate: bool = False,
                 **kwargs) -> Iterable[DenseConfiguration]:
        """Maximize acquisition function using ``_maximize``.

        Parameters
        ----------
        trials: ~xbbo.utils.history_container.Trials
            trials object
        num_points: int
            number of points to be sampled
        random_configuration_chooser: ~xbbo.acq_maximizer.random_configuration_chooser.RandomConfigurationChooser
            part of the returned ChallengerList such
            that we can interleave random configurations
            by a scheme defined by the random_configuration_chooser;
            random_configuration_chooser.next_smbo_iteration()
            is called at the end of this function
        **kwargs
            passed to acquisition function

        Returns
        -------
        Iterable[DenseConfiguration]
            to be concrete: ~xbbo.ei_optimization.ChallengerList
        """

        next_configs_by_local_search = self.local_search._maximize(
            trials, self.n_sls_iterations, **kwargs)

        # Get configurations sorted by EI
        new_kwargs = {"_sorted":True}
        new_kwargs.update(kwargs)
        next_configs_by_random_search_sorted = self.random_search._maximize(
            trials,
            num_points - len(next_configs_by_local_search),
            **new_kwargs)

        # Having the configurations from random search, sorted by their
        # acquisition function value is important for the first few iterations
        # of xbbo. As long as the random forest predicts constant value, we
        # want to use only random configurations. Having them at the begging of
        # the list ensures this (even after adding the configurations by local
        # search, and then sorting them)
        next_configs_by_acq_value = (next_configs_by_random_search_sorted +
                                     next_configs_by_local_search)
        next_configs_by_acq_value.sort(reverse=True, key=lambda x: x[0])
        logger.debug(
            "First 10 acq func (origin) values of selected configurations: %s",
            str([[_[0], _[1].origin] for _ in next_configs_by_acq_value[:10]]))
        configs = [_[1] for _ in next_configs_by_acq_value]
        return self.unique(configs=configs) if drop_self_duplicate else configs

    def _maximize(self, trials: Trials, num_points: int,
                  **kwargs) -> Iterable[Tuple[float, DenseConfiguration]]:
        raise NotImplementedError()



    # """Get candidate solutions via random sampling of configurations.

    # Parameters
    # ----------
    # acquisition_function : ~xbbo.optimizer.acquisition.AbstractAcquisitionFunction

    # config_space : ~xbbo.configspace.DenseConfigurationSpace

    # rng : np.random.RandomState or int, optional
    # """
    # def __init__(
    #     self,
    #     acquisition_function: AbstractAcquisitionFunction,
    #     config_space: DenseConfigurationSpace,
    #     rng: np.random.RandomState = np.random.RandomState(42),
    #     design_method: Optional[str] = 'sobol',
    # ):
    #     super().__init__(acquisition_function, config_space, rng)
    #     self.design_method = design_method
    #     types, bounds = get_types(self.config_space)
    #     self.bounds = bounds
    #     self.dim = len(self.bounds)

    # def _maximize(self,
    #               trials: Trials,
    #               num_points: int,
    #               _sorted: bool = False,
    #               **kwargs) -> List[Tuple[float, DenseConfiguration]]:
    #     """Design sampled configurations

    #     Parameters
    #     ----------
    #     trials: ~xbbo.trials.trials.trials
    #         trials object
    #     stats: ~xbbo.stats.stats.Stats
    #         current stats object
    #     num_points: int
    #         number of points to be sampled
    #     _sorted: bool
    #         whether random configurations are sorted according to acquisition function

    #     Returns
    #     -------
    #     iterable
    #         An iterable consistng of
    #         tuple(acqusition_value, :class:`xbbo.configspace.DenseConfiguration`).
    #     """
    #     X = trials.get_array()
    #     fX = np.asarray(trials._his_observe_value)
    #     length_scale = kwargs.get('length_scale', np.ones(self.dim))
    #     length = kwargs['length']
    #     weights = np.array(length_scale)
    #     # Create the trust region boundaries
    #     x_center = X[fX.argmin().item(), :][None, :]
    #     weights = weights / weights.mean(
    #     )  # This will make the next line more stable
    #     weights = weights / np.prod(np.power(
    #         weights, 1.0 / len(weights)))  # We now have weights.prod() = 1
    #     lb = np.clip(x_center - weights * length / 2.0, 0.0, 1.0)
    #     ub = np.clip(x_center + weights * length / 2.0, 0.0, 1.0)

    #     # Draw a Sobolev sequence in [lb, ub]
    #     sobol_gen = Sobol(d=self.dim,
    #                       scramble=True,
    #                       seed=self.rng.randint(MAXINT))
    #     pert = sobol_gen.random(num_points)
    #     pert = lb + (ub - lb) * pert

    #     # Create a perturbation mask
    #     prob_perturb = min(20.0 / self.dim, 1.0)
    #     mask = self.rng.rand(num_points,
    #                          self.dim) <= prob_perturb
    #     ind = np.where(np.sum(mask, axis=1) == 0)[0]
    #     mask[ind,
    #          self.rng.randint(0, self.dim - 1, size=len(ind))] = 1

    #     # Create candidate points
    #     X_cand = x_center.copy() * np.ones(
    #         (num_points, self.dim))
    #     X_cand[mask] = pert[mask]
        
    #     acq_and_configs = []
    #     for n in num_points:
    #         config = DenseConfiguration.from_array(self.config_space, X_cand, use_dense=trials.use_dense)
    #         y_cand = self.acquisition_function(config)
    #         acq_and_configs.append((y_cand, config))


# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

import logging
import typing

import numpy as np
from pyrfr import regression
from xbbo.configspace.space import DenseConfigurationSpace
from xbbo.surrogate.base import BaseRF
from xbbo.utils.constants import MAXINT
from xbbo.utils.util import get_types


class RandomForestWithInstances(BaseRF):
    """Random forest that takes instance features into account.

    Attributes
    ----------
    rf_opts : regression.rf_opts
        Random forest hyperparameter
    n_points_per_tree : int
    rf : regression.binary_rss_forest
        Only available after training
    hypers: list
        List of random forest hyperparameters
    unlog_y: bool
    seed : int
    types : np.ndarray
    bounds : list
    rng : np.random.RandomState
    logger : logging.logger
    """
    def __init__(self,
                 configspace: DenseConfigurationSpace,
                 log_y: bool = False,
                 num_trees: int = 10,
                 do_bootstrapping: bool = True,
                 n_points_per_tree: int = -1,
                 ratio_features: float = 5. / 6.,
                 min_samples_split: int = 3,
                 min_samples_leaf: int = 3,
                 max_depth: int = 2**20,
                 eps_purity: float = 1e-8,
                 max_num_nodes: int = 2**20,
                 rng: np.random.RandomState=np.random.RandomState(42),
                 **kwargs):
        """
        Parameters
        ----------
        types : np.ndarray (D)
            Specifies the number of categorical values of an input dimension where
            the i-th entry corresponds to the i-th input dimension. Let's say we
            have 2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass np.array([2, 0]). Note that we count starting from 0.
        bounds : list
            Specifies the bounds for continuous features.
        log_y: bool
            y values (passed to this RF) are expected to be log(y) transformed;
            this will be considered during predicting
        num_trees : int
            The number of trees in the random forest.
        do_bootstrapping : bool
            Turns on / off bootstrapping in the random forest.
        n_points_per_tree : int
            Number of points per tree. If <= 0 X.shape[0] will be used
            in _train(X, y) instead
        ratio_features : float
            The ratio of features that are considered for splitting.
        min_samples_split : int
            The minimum number of data points to perform a split.
        min_samples_leaf : int
            The minimum number of data points in a leaf.
        max_depth : int
            The maximum depth of a single tree.
        eps_purity : float
            The minimum difference between two target values to be considered
            different
        max_num_nodes : int
            The maxmimum total number of nodes in a tree
        seed : int
            The seed that is passed to the random_forest_run library.
        """
        types, bounds = get_types(configspace)
        super().__init__(configspace, types, bounds, **kwargs)

        self.log_y = log_y
        seed = rng.randint(MAXINT)
        self.rng = regression.default_random_engine(seed)

        self.rf_opts = regression.forest_opts()
        self.rf_opts.num_trees = num_trees
        self.rf_opts.do_bootstrapping = do_bootstrapping
        max_features = 0 if ratio_features > 1.0 else \
            max(1, int(types.shape[0] * ratio_features))
        self.rf_opts.tree_opts.max_features = max_features
        self.rf_opts.tree_opts.min_samples_to_split = min_samples_split
        self.rf_opts.tree_opts.min_samples_in_leaf = min_samples_leaf
        self.rf_opts.tree_opts.max_depth = max_depth
        self.rf_opts.tree_opts.epsilon_purity = eps_purity
        self.rf_opts.tree_opts.max_num_nodes = max_num_nodes
        self.rf_opts.compute_law_of_total_variance = False

        self.n_points_per_tree = n_points_per_tree
        self.rf = None  # type: regression.binary_rss_forest

        # This list well be read out by save_iteration() in the solver
        self.hypers = [
            num_trees, max_num_nodes, do_bootstrapping, n_points_per_tree,
            ratio_features, min_samples_split, min_samples_leaf, max_depth,
            eps_purity, seed
        ]
        self.seed = seed

        self.logger = logging.getLogger(self.__module__ + "." +
                                        self.__class__.__name__)

    def _train(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """Trains the random forest on X and y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, ]
            The corresponding target values.

        Returns
        -------
        self
        """
        X = self._impute_inactive(X)
        self.X = X
        self.y = y.flatten()

        if self.n_points_per_tree <= 0:
            self.rf_opts.num_data_points_per_tree = self.X.shape[0]
        else:
            self.rf_opts.num_data_points_per_tree = self.n_points_per_tree
        self.rf = regression.binary_rss_forest()
        self.rf.options = self.rf_opts
        data = self._init_data_container(self.X, self.y)
        self.rf.fit(data, rng=self.rng)
        return self

    def _init_data_container(self, X: np.ndarray, y: np.ndarray):
        """Fills a pyrfr default data container, s.t. the forest knows
        categoricals and bounds for continous data

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features]
            Input data points
        y : np.ndarray [n_samples, ]
            Corresponding target values

        Returns
        -------
        data : regression.default_data_container
            The filled data container that pyrfr can interpret
        """
        # retrieve the types and the bounds from the ConfigSpace
        data = regression.default_data_container(X.shape[1])

        for i, (mn, mx) in enumerate(self.bounds):
            if np.isnan(mx):
                data.set_type_of_feature(i, mn)
            else:
                data.set_bounds_of_feature(i, mn, mx)

        for row_X, row_y in zip(X, y):
            data.add_data_point(row_X, row_y)
        return data

    def _predict(self, X: np.ndarray,**kwargs) -> typing.Tuple[np.ndarray, np.ndarray]:
        """Predict means and variances for given X.

        Parameters
        ----------
        X : np.ndarray of shape = [n_samples,
                                   n_features (config + instance features)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, 1]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, 1]
            Predictive variance
        """
        if len(X.shape) != 2:
            raise ValueError('Expected 2d array, got %dd array!' %
                             len(X.shape))
        if X.shape[1] != self.types.shape[0]:
            raise ValueError('Rows in X should have %d entries but have %d!' %
                             (self.types.shape[0], X.shape[1]))
        X = self._impute_inactive(X)
        means, vars_ = [], []
        for row_X in X:
            if self.log_y:
                preds_per_tree = self.rf.all_leaf_values(row_X)
                means_per_tree = []
                for preds in preds_per_tree:
                    # within one tree, we want to use the
                    # arithmetic mean and not the geometric mean
                    means_per_tree.append(np.log(np.mean(np.exp(preds))))
                mean = np.mean(means_per_tree)
                var = np.var(means_per_tree
                             )  # variance over trees as uncertainty estimate
            else:
                mean, var = self.rf.predict_mean_var(row_X)
            means.append(mean)
            vars_.append(var)
        means = np.array(means)
        vars_ = np.array(vars_)

        return means.reshape((-1, 1)), vars_.reshape((-1, 1))

    def predict_marginalized_over_instances(self, X: np.ndarray):
        """Predict mean and variance marginalized over all instances.

        Returns the predictive mean and variance marginalised over all
        instances for a set of configurations.

        Note
        ----
        This method overwrites the same method of ~smac.epm.base_epm.AbstractEPM;
        the following method is random forest specific
        and follows the SMAC2 implementation;
        it requires no distribution assumption
        to marginalize the uncertainty estimates

        Parameters
        ----------
        X : np.ndarray
            [n_samples, n_features (config)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, 1]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, 1]
            Predictive variance
        """

        if self.instance_features is None or \
                len(self.instance_features) == 0:
            mean, var = self.predict(X)
            var[var < self.var_threshold] = self.var_threshold
            var[np.isnan(var)] = self.var_threshold
            return mean, var

        if len(X.shape) != 2:
            raise ValueError('Expected 2d array, got %dd array!' %
                             len(X.shape))
        if X.shape[1] != self.bounds.shape[0]:
            raise ValueError('Rows in X should have %d entries but have %d!' %
                             (self.bounds.shape[0], X.shape[1]))
        X = self._impute_inactive(X)
        mean = np.zeros(X.shape[0])
        var = np.zeros(X.shape[0])
        for i, x in enumerate(X):

            # marginalize over instances
            # 1. get all leaf values for each tree
            preds_trees = [[] for i in range(self.rf_opts.num_trees)]

            for feat in self.instance_features:
                x_ = np.concatenate([x, feat])
                preds_per_tree = self.rf.all_leaf_values(x_)
                for tree_id, preds in enumerate(preds_per_tree):
                    preds_trees[tree_id] += preds

            # 2. average in each tree
            for tree_id in range(self.rf_opts.num_trees):
                if self.log_y:
                    preds_trees[tree_id] = \
                        np.log(np.mean(np.exp(preds_trees[tree_id])))
                else:
                    preds_trees[tree_id] = np.mean(preds_trees[tree_id])

            # 3. compute statistics across trees
            mean_x = np.mean(preds_trees)
            var_x = np.var(preds_trees)
            if var_x < self.var_threshold:
                var_x = self.var_threshold

            var[i] = var_x
            mean[i] = mean_x

        if len(mean.shape) == 1:
            mean = mean.reshape((-1, 1))
        if len(var.shape) == 1:
            var = var.reshape((-1, 1))

        return mean, var
